/*
 *             High performance packet classification 
 *                     <http://www.hipac.org>
 *
 * (c) 2004-2005   MARA Systems AB <http://www.marasystems.com>
 *                 +-----------------------------+
 *                 |       Michael Bellion       |
 *                 |  <michael@marasystems.com>  |
 *                 +-----------------------------+
 *
 * (c) 2002-2003   hipac core team <nf@hipac.org>:
 *     +---------------------------+--------------------------+
 *     |      Michael Bellion      |       Thomas Heinz       |
 *     |   <mbellion@hipac.org>    |   <creatix@hipac.org>    |
 *     +---------------------------+--------------------------+
 *
 * Licenced under the GNU General Public Licence, version 2.
 */


#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <sys/time.h>
#include "libnfhipac.h"
#include "hipac.h"
#include "nfhp_com.h"

#ifndef PROC_SYS_MODPROBE
#  define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
#endif

#define MIN(a, b)          ((a) <= (b) ? (a) : (b))
#define IS_LIST_CMD(cmd)   ((cmd) == CMD_LIST)
#define CMD_SIZE     4096
#define RECV_SIZE   65536

struct nfhp_cmd *
nlhp_new_cmd(int cmd, int *size)
{
	struct nlmsghdr *msg;
	struct nfhp_cmd *c;

	if (cmd > CMD_MAX) {
		return NULL;
	}
	
	msg = calloc(1, CMD_SIZE);
	if (msg == NULL) {
		return NULL;
	}
	if (size != NULL){
		*size = CMD_SIZE - NLMSG_LENGTH(0);
	}
	c = NLMSG_DATA(msg);
	c->cmd = cmd;
	return c;
}

void
nlhp_free_cmd(struct nfhp_cmd *cmd)
{
	if (cmd == NULL) {
		return;
	}
	free((char *) cmd - NLMSG_LENGTH(0));
}

struct nfhp_cmd *
nlhp_copy_cmd(struct nfhp_cmd *cmd, int size)
{
	struct nlmsghdr *clone;

	if (cmd == NULL || size < sizeof(*cmd)) {
		return NULL;
	}
	clone = malloc(NLMSG_SPACE(size));
	if (clone == NULL) {
		return NULL;
	}
	memcpy(clone, (char *) cmd - NLMSG_LENGTH(0), NLMSG_SPACE(size));
	return NLMSG_DATA(clone);
}

/* read location of modprobe binary from procfs */
static char *
get_modprobe_cmd(void)
{
	int procfile, len;
	char *ret;
		
	procfile = open(PROC_SYS_MODPROBE, O_RDONLY);
	if (procfile < 0) {
		return NULL;
	}
	
	ret = malloc(1024);
	if (ret == NULL) {
		goto error;
	}
	switch (read(procfile, ret, 1024)) {
	    case -1:   goto error;
	    case 1024: goto error;
	}
	len = strlen(ret);
	if (ret[len - 1] == '\n') {
		ret[len - 1] = '\0';
	}
	close(procfile);
	return ret;
 
 error:
	free(ret);
	close(procfile);
	return NULL;
}

/* try to load module specified by modname */
static int
modprobe(char *modname)
{
	char *modprobe_cmd = NULL;
	char *argv[3];
	
	modprobe_cmd = get_modprobe_cmd();
	if (modprobe_cmd == NULL) {
		return -1;
	}
	switch (fork()) {
	    case 0:
		    argv[0] = modprobe_cmd;
		    argv[1] = modname;
		    argv[2] = NULL;
		    execv(argv[0], argv);
		    /* usually not reached */
		    exit(0);

	    case -1:
		    return -1;
		    
	    default: /* parent */
		    wait(NULL);
	}
	free(modprobe_cmd);
	return 0;
}

static inline int
rcv_kern(int sock, void *buf, int len, struct sockaddr_nl *addr,
	 socklen_t *addrlen)
{
	fd_set set;
	int stat;
	
	FD_SET(sock, &set);
	stat = select(sock + 1, &set, NULL, NULL, NULL);
	FD_ZERO(&set);
	if (stat == 0) {
		return -2;
	} else if (stat == -1) {
		return -1;
	}
	return recvfrom(sock, buf, len, 0, (struct sockaddr *) addr, addrlen);
}

static inline int
check_reply(struct nlmsghdr *msg, int size, int min_size)
{
	int pl;

	if (size < 0) {
		return size;
	}
	if (size <= NLMSG_LENGTH(0)) {
		fprintf(stderr, "internal error: reply has empty payload\n");
		return -5;
	}
	pl = NLMSG_PAYLOAD(msg, 0);
	if (pl < min_size) {
		fprintf(stderr, "internal error: payload too short\n");
		return -5;
	}
	if (!NLMSG_OK(msg, size)) {
		fprintf(stderr, "internal error: reply broken\n");
		return -5;
	}
	return pl;
}

int
nlhp_send_cmd(struct nfhp_cmd *cmd, int cmd_size,
	      int (*process_list) (const void *data, int len),
	      int *err)
{
	int rcvbuf_size;
	struct sockaddr_nl addr = {0};
	socklen_t addrlen = sizeof(addr);
	struct nlmsghdr *msg, *reply;
	int sock, stat, pl;
	void *data;

	if (cmd == NULL || err == NULL || cmd_size < sizeof(*cmd) ||
	    (IS_LIST_CMD(cmd->cmd) && process_list == NULL)) {
		return -5;
	}

	stat = 0;
	*err = 0;

	/* open socket */
	if ((sock = socket(PF_NETLINK, SOCK_RAW, NLHP_PROTO)) == -1) {
		return -1;
	}
	addr.nl_family = AF_NETLINK;
	if (bind(sock, (struct sockaddr *) &addr, addrlen) == -1) {
		stat = -1;
		goto end;
	}
	
	/* fill netlink header */
	msg = (struct nlmsghdr *) ((char *) cmd - NLMSG_LENGTH(0));
	msg->nlmsg_len   = NLMSG_LENGTH(cmd_size);
	msg->nlmsg_type  = NLHP_TYPE;
	msg->nlmsg_flags = NLM_F_REQUEST | NLM_F_ACK;
	msg->nlmsg_pid   = getpid();
	if (!NLMSG_OK(msg, NLMSG_LENGTH(cmd_size))) {
		fprintf(stderr, "internal error: message built failed\n");
		stat = -5;
		goto end;
	}

	/* send request */
	if (sendto(sock, msg, NLMSG_SPACE(cmd_size), 0,
		   (struct sockaddr *) &addr, addrlen) == -1) {
		modprobe("nf_hipac");
		if (sendto(sock, msg, NLMSG_SPACE(cmd_size),
			   0, (struct sockaddr *) &addr, addrlen) == -1) {
			stat = -1;
			goto end;
		}
	}

	if (!IS_LIST_CMD(cmd->cmd))
		rcvbuf_size = NLMSG_SPACE(sizeof(int32_t));
	else rcvbuf_size = RECV_SIZE;
	
	/* receive reply */
       	reply = malloc(rcvbuf_size);
	if (reply == NULL) {
		stat = -3;
		goto end;
	}

	/* non-list command */
	if (!IS_LIST_CMD(cmd->cmd)) {
		int size;
		size = rcv_kern(sock, reply, rcvbuf_size, &addr, &addrlen);
		pl = check_reply(reply, size, sizeof(int32_t));
		if (pl < 0) {
			stat = pl;
			goto end_free;
		}
		if (reply->nlmsg_type == NLMSG_ERROR) {
			*err = *(int32_t *) NLMSG_DATA(reply);
			goto end_free;
		} else {
			fprintf(stderr, "internal error: netlink error\n");
			stat = -5;
			goto end_free;
		}
	} else {
		/* list callback */
		while (1) {
			int size;
			size = rcv_kern(sock, reply, rcvbuf_size, 
					&addr, &addrlen);
			data = NLMSG_DATA(reply);
			pl = check_reply(reply, size, sizeof(int32_t));
			if (pl < 0) {
				stat = pl;
				break;
			}
			if (pl == sizeof(int32_t) &&
			    (reply->nlmsg_type == NLMSG_DONE ||
			     reply->nlmsg_type == NLMSG_ERROR)) {
				if (*(int32_t *) data < 0) {
					/* listing ends abnormally */
					*err = *(int32_t *) data;
				} else {
					*err = 0;
				}
				break;
			}
			if (pl < MIN(sizeof(struct nfhp_list_chain),
				     sizeof(struct nfhp_list_rule))) {
				fprintf(stderr, "internal error: "
					"payload too short\n");
				stat = -5;
				break;
			}
			if (process_list(data, pl) < 0) {
				stat = -4;
				break;
			}
			if (reply->nlmsg_type == NLMSG_DONE) {
				*err = 0;
				break;
			}
		}
	}

 end_free:
	free(reply);
 end:
	close(sock);
	return stat;
}

const char *
nlhp_error(int err)
{
	static const char *errmsg[] =
	{
		/* hipac_error */
		[-HE_OK]                        = "Operation successful",
		[-HE_IMPOSSIBLE_CONDITION]      = "Impossible condition (e.g. illegal function call)",
		[-HE_LOW_MEMORY]                = "Not enough memory available",
		[-HE_CHAIN_EXISTS]              = "Chain exists already",
		[-HE_CHAIN_NOT_EXISTENT]        = "Chain does not exist",
		[-HE_CHAIN_IS_EMPTY]            = "Chain is empty",
		[-HE_CHAIN_NOT_EMPTY]           = "Chain is not empty",
		[-HE_CHAIN_IS_USERDEFINED]      = "Chain is user defined",
		[-HE_CHAIN_IS_CONNECTED]        = "Chain is connected",
		[-HE_CHAIN_IS_REFERENCED]       = "Chain is still referenced",
		[-HE_CHAIN_NOT_NATIVE]          = "Chain is not native",
		[-HE_CHAIN_IS_NATIVE]           = "Chain is native",
		[-HE_RULE_NOT_EXISTENT]         = "Rule does not exist",
		[-HE_RULE_ORIGIN_MISMATCH]      = "Rule origin mismatch",
		[-HE_RULE_PREFIX_MISMATCH]      = "Rule prefix mismatch",
		[-HE_LOOP_DETECTED]             = "Loop detected",
		[-HE_REC_LIMIT]                 = "Chain depth limit reached",
		[-HE_TARGET_CHAIN_NOT_EXISTENT] = "Target chain does not exist",
		[-HE_TARGET_CHAIN_IS_NATIVE]    = "Target chain is native",
		[-HE_NATIVE_CHAIN_EXISTS]       = "Native chain exists",
		
		/* nfhipac_error */
		[-NFHE_INDEX]   = "Unable to retrieve unused ifindex",
		[-NFHE_NOMSG]   = "Incorrect message format",
		[-NFHE_CMD]     = "Invalid command",
		[-NFHE_LABEL]   = "Empty chain label",
		[-NFHE_NLABEL]  = "Empty new chain label",
		[-NFHE_POLICY]  = "Invalid policy",
		[-NFHE_ACTION]  = "Invalid action",
		[-NFHE_NMCT]    = "Invalid native match count",
		[-NFHE_IEOFF]   = "Invalid target_offset/next_offset in ipt_entry",
		[-NFHE_SORT]    = "Native matches not sorted or dimid duplicate",
		[-NFHE_MINT]    = "Invalid interval in native match",
		[-NFHE_DEVA]    = "Native interface match but no corresponding interface name",
		[-NFHE_DEVB]    = "Interface name but no corresponding native interface match",
		[-NFHE_FRAG]    = "Invalid fragment match",
		[-NFHE_PROTO]   = "Invalid protocol match",
		[-NFHE_SYN]     = "Invalid syn match",
		[-NFHE_STATE]   = "Invalid state match",
		[-NFHE_TCP]     = "tcp match dependency failure",
		[-NFHE_TCPUDP]  = "tcp or udp match dependency failure",
		[-NFHE_ICMP]    = "icmp match dependency failure",
		[-NFHE_CMPMIS]  = "Missing cmp_len array",
		[-NFHE_CMPSH]   = "cmp_len array too short",
		[-NFHE_CMPLA]   = "cmp_len array contains a value larger than the corresponding ipt match/target size",
		[-NFHE_ORIGIN]  = "Illegal combination of matches (no valid origin)",
		[-NFHE_IPTMSZ]  = "Invalid ipt match size",
		[-NFHE_IPTMCH]  = "checkentry fails for ipt match",
		[-NFHE_IPTTMI]  = "Missing ipt target",
		[-NFHE_IPTTSZ]  = "Invalid ipt target size",
		[-NFHE_IPTTCH]  = "checkentry fails for ipt target",
		[-NFHE_TOFF]    = "Invalid target_offset",
		[-NFHE_CHAINE]  = "Empty chain name",
		[-NFHE_CHAINL]  = "Chain name too long",
		[-NFHE_CT]      = "Kernel does not have support for connection tracking, please recompile",
		[-NFHE_CTHELP]  = "Unable to load connection tracking helper module (nf_hipac_cthelp.o)",
		[-NFHE_ILL]     = "Illegal condition",
		[-NFHE_IMPL]    = "Feature not yet implemented"
	};
	static char buf[30];
	int32_t n;

	if (err <= NFHE_SYSOFF) {
		return strerror(NFHE_TO_ERRNO(err));
	}

	n = -err;
	if (n < 0 || n >= sizeof(errmsg) / sizeof(*errmsg) ||
	    errmsg[n] == NULL) {
		snprintf(buf, sizeof(buf), "Unkown error %d\n", err);
		return buf;
	}
	return errmsg[n];
}
