/*
 *             High performance packet classification 
 *                     <http://www.hipac.org>
 *
 * (c) 2004-2005   MARA Systems AB <http://www.marasystems.com>
 *                 +-----------------------------+
 *                 |       Michael Bellion       |
 *                 |  <michael@marasystems.com>  |
 *                 +-----------------------------+
 *
 * (c) 2002-2003   hipac core team <nf@hipac.org>:
 *     +---------------------------+--------------------------+
 *     |      Michael Bellion      |       Thomas Heinz       |
 *     |   <mbellion@hipac.org>    |   <creatix@hipac.org>    |
 *     +---------------------------+--------------------------+
 *
 * heavily based on iptables code from the netfilter project
 * which is (C) by the netfilter coreteam <coreteam@netfilter.org>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>	
#include <getopt.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <link.h>
#include <dlfcn.h>
#include <sys/utsname.h>


#include <features.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#  include <net/if.h>
#  include <netinet/in.h>
#  include <asm/types.h>
#else /* libc5 */
#  include <linux/if.h>
#  include <linux/in.h>
#  include <linux/types.h>
#endif

#include <linux/netfilter_ipv4/ip_tables.h>

#include "nfhp_com.h"
#include "libnfhipac.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

static int listing_start = 0;

#define OPT_NONE	0x00000U
#define OPT_NUMERIC	0x00001U
#define OPT_VERBOSE	0x00002U
#define OPT_LINENUMBERS 0x00004U
#define OPT_DEBUG       0x00008U
#define OPT_DEBUGN      0x00010U
#define OPT_JUMP        0x00020U

#define NUMBER_OF_OPT	6

static const char optflags[NUMBER_OF_OPT + NUMBER_OF_DIM][18]
= {
	[0] = "n",
	[1] = "v",
	[2] = "-line",
	[3] = "-debug",
	[4] = "-debugn",
	[5] = "j",
	[NUMBER_OF_OPT + DIMID_STATE] = "-state",
	[NUMBER_OF_OPT + DIMID_SRC_IP ] = "s",
	[NUMBER_OF_OPT + DIMID_DEST_IP ] = "d",
	[NUMBER_OF_OPT + DIMID_INIFACE] = "i",
	[NUMBER_OF_OPT + DIMID_OUTIFACE] = "o",
	[NUMBER_OF_OPT + DIMID_PROTO] = "p",
	[NUMBER_OF_OPT + DIMID_FRAGMENT] = "f",
	[NUMBER_OF_OPT + DIMID_DPORT] = "-dport",
	[NUMBER_OF_OPT + DIMID_SPORT] = "-sport",
	[NUMBER_OF_OPT + DIMID_SYN] = "-syn or --not-syn",
	[NUMBER_OF_OPT + DIMID_ICMP_TYPE] = "-icmp-type",
	[NUMBER_OF_OPT + DIMID_TTL] = "-ttl",
};



#define HIPAC_NATIVE_MATCHES 5
static const char native_matches[HIPAC_NATIVE_MATCHES][6]
= { "tcp", "udp", "icmp", "state", "ttl"};


static const char cmdflags[] = { ' ', 'A', 'I', 'D', 'D', 'R', 'F',
				 'N', 'X', 'E', 'P', 'L'};



static char commands_v_options[CMD_MAX + 1][2] = {
	//type 0: -n, --line, --exact
	//type 1: -s, -p, -sport, -dport, ...
        //      0   1   2  

/*NONE*/         {'x','x'},
/*APPEND*/       {'x',' '},
/*INSERT*/       {'x',' '},
/*DELETE_RULE*/  {'x',' '},
/*DELETE_POS*/   {'x','x'},
/*REPLACE*/      {'x',' '},
/*FLUSH*/        {'x','x'},
/*NEW_CHAIN*/    {'x','x'},
/*DELETE_CHAIN*/ {'x','x'},
/*RENAME_CHAIN*/ {'x','x'},
/*SET_POLICY*/   {'x','x'},
/*LIST*/         {' ','x'},
};


static struct option original_opts[] = {
	{ "append", 1, 0, 'A' },
	{ "delete", 1, 0,  'D' },
	{ "insert", 1, 0,  'I' },
	{ "replace", 1, 0,  'R' },
	{ "list", 2, 0,  'L' },
	{ "flush", 2, 0,  'F' },
	{ "new-chain", 1, 0,  'N' },
	{ "delete-chain", 2, 0,  'X' },
	{ "rename-chain", 1, 0,  'E' },
	{ "policy", 1, 0,  'P' },
	{ "source", 1, 0, 's' },
	{ "destination", 1, 0,  'd' },
	{ "src", 1, 0,  's' }, /* synonym */
	{ "dst", 1, 0,  'd' }, /* synonym */
	{ "protocol", 1, 0,  'p' },
	{ "in-interface", 1, 0, 'i' },
	{ "jump", 1, 0, 'j' },
	{ "table", 1, 0, 't' },
	{ "match", 1, 0, 'm' },
	{ "numeric", 0, 0, 'n' },
	{ "out-interface", 1, 0, 'o' },
	{ "verbose", 0, 0, 'v' },
	{ "fragments", 0, 0, 'f' },
	{ "version", 0, 0, 'V' },
	{ "help", 2, 0, 'h' },
	{ "debug", 0, 0, 9 },
	{ "debugn", 0, 0, 10 },
	{ "line-numbers", 0, 0, 11 },
	{ "state", 1, 0, 12 },
	{ "ttl", 1, 0, 14 },
	{ "icmp-type", 1, 0, 15 },	
	{ "source-port", 1, 0, 16 },
	{ "sport", 1, 0, 16 }, /* synonym */
	{ "destination-port", 1, 0, 17 },
	{ "dport", 1, 0, 17 }, /* synonym */
	{ "syn", 0, 0, 18 },
	{ "not-syn", 0, 0, 19 },
	{ 0 }
};

/* we need this for iptables-restore.  iptables-restore.c sets line to the
 * current line of the input file, in order  to give a more precise error
 * message.  iptables itself doesn't need this, so it is initialized to the
 * magic number of -1 */
int line = -1;

static struct option *opts = original_opts;
static unsigned int global_option_offset = 0;
static unsigned int options = 0;

/* A few hardcoded protocols for 'all' and in case the user has no
   /etc/protocols */
struct pprot {
	char *name;
	u_int8_t num;
};

/* Primitive headers... */
/* defined in netinet/in.h */
#if 0
#ifndef IPPROTO_ESP
#define IPPROTO_ESP 50
#endif
#ifndef IPPROTO_AH
#define IPPROTO_AH 51
#endif
#endif

static const struct pprot chain_protos[] = {
	{ "tcp", IPPROTO_TCP },
	{ "udp", IPPROTO_UDP },
	{ "icmp", IPPROTO_ICMP },
	{ "esp", IPPROTO_ESP },
	{ "ah", IPPROTO_AH },
	{ "sctp", IPPROTO_SCTP },
	{ "all", 0 },
};

struct icmp_names {
	const char *name;
	u_int8_t type;
	u_int8_t code_min, code_max;
};

static const struct icmp_names icmp_codes[] = {
	{ "echo-reply", 0, 0, 0xFF },
	/* Alias */ { "pong", 0, 0, 0xFF },

	{ "destination-unreachable", 3, 0, 0xFF },
	{   "network-unreachable", 3, 0, 0 },
	{   "host-unreachable", 3, 1, 1 },
	{   "protocol-unreachable", 3, 2, 2 },
	{   "port-unreachable", 3, 3, 3 },
	{   "fragmentation-needed", 3, 4, 4 },
	{   "source-route-failed", 3, 5, 5 },
	{   "network-unknown", 3, 6, 6 },
	{   "host-unknown", 3, 7, 7 },
	{   "network-prohibited", 3, 9, 9 },
	{   "host-prohibited", 3, 10, 10 },
	{   "TOS-network-unreachable", 3, 11, 11 },
	{   "TOS-host-unreachable", 3, 12, 12 },
	{   "communication-prohibited", 3, 13, 13 },
	{   "host-precedence-violation", 3, 14, 14 },
	{   "precedence-cutoff", 3, 15, 15 },

	{ "source-quench", 4, 0, 0xFF },

	{ "redirect", 5, 0, 0xFF },
	{   "network-redirect", 5, 0, 0 },
	{   "host-redirect", 5, 1, 1 },
	{   "TOS-network-redirect", 5, 2, 2 },
	{   "TOS-host-redirect", 5, 3, 3 },

	{ "echo-request", 8, 0, 0xFF },
	/* Alias */ { "ping", 8, 0, 0xFF },

	{ "router-advertisement", 9, 0, 0xFF },

	{ "router-solicitation", 10, 0, 0xFF },

	{ "time-exceeded", 11, 0, 0xFF },
	/* Alias */ { "ttl-exceeded", 11, 0, 0xFF },
	{   "ttl-zero-during-transit", 11, 0, 0 },
	{   "ttl-zero-during-reassembly", 11, 1, 1 },

	{ "parameter-problem", 12, 0, 0xFF },
	{   "ip-header-bad", 12, 0, 0 },
	{   "required-option-missing", 12, 1, 1 },

	{ "timestamp-request", 13, 0, 0xFF },

	{ "timestamp-reply", 14, 0, 0xFF },

	{ "address-mask-request", 17, 0, 0xFF },

	{ "address-mask-reply", 18, 0, 0xFF }
};

struct dim_match
{
	u_int8_t invert;
	u_int32_t left;
        u_int32_t right;
};

#define OPTION_OFFSET 256

#define FMT_NUMERIC     0x0001
#define FMT_NOCOUNTS    0x0002
#define FMT_KILOMEGAGIGA 0x0004
#define FMT_OPTIONS     0x0008
#define FMT_NOTABLE     0x0010
#define FMT_NOTARGET    0x0020
#define FMT_VIA         0x0040
#define FMT_NONEWLINE   0x0080
#define FMT_LINENUMBERS 0x0100

#define FMT_PRINT_RULE (FMT_NOCOUNTS | FMT_OPTIONS | FMT_VIA \
                        | FMT_NUMERIC | FMT_NOTABLE)


/*
 * =============== iptables matches/targets compatibility ===============
 */

#ifndef PROC_SYS_MODPROBE
#define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
#endif


/*
 * libiptc.h declarations
 */

#ifndef IPT_MIN_ALIGN
/* ipt_entry has pointers and u_int64_t's in it, so if you align to
   it, you'll also align to any crazy matches and targets someone
   might write */
#define IPT_MIN_ALIGN (__alignof__(struct ipt_entry))
#endif

#define IPT_ALIGN(s) (((s) + ((IPT_MIN_ALIGN)-1)) & ~((IPT_MIN_ALIGN)-1))

typedef char ipt_chainlabel[32];

#define IPTC_LABEL_ACCEPT  "ACCEPT"
#define IPTC_LABEL_DROP    "DROP"
#define IPTC_LABEL_QUEUE   "QUEUE"
#define IPTC_LABEL_RETURN  "RETURN"

/* Transparent handle type. */
typedef struct iptc_handle *iptc_handle_t;


/*
 * iptables.h declarations
 */

#ifndef IPT_LIB_DIR
#define IPT_LIB_DIR "/usr/local/lib/iptables"
#endif

#ifndef IPPROTO_SCTP
#define IPPROTO_SCTP 132
#endif

#ifndef IPT_SO_GET_REVISION_MATCH /* Old kernel source. */
#define IPT_SO_GET_REVISION_MATCH       (IPT_BASE_CTL + 2)
#define IPT_SO_GET_REVISION_TARGET      (IPT_BASE_CTL + 3)

struct ipt_get_revision
{
	char name[IPT_FUNCTION_MAXNAMELEN-1];
	
	u_int8_t revision;
};
#endif /* IPT_SO_GET_REVISION_MATCH   Old kernel source */
  	 
struct iptables_rule_match
{
	struct iptables_rule_match *next;
	
	struct iptables_match *match;
};

struct iptables_match
{
	struct iptables_match *next;

	ipt_chainlabel name;
	
#ifndef IPTABLES_1_2
	/* Revision of match (0 by default). */
	u_int8_t revision;
#endif

	const char *version;

	/* Size of match data. */
	size_t size;

	/* Size of match data relevent for userspace comparison purposes */
	size_t userspacesize;

	/* Function which prints out usage message. */
	void (*help)(void);

	/* Initialize the match. */
	void (*init)(struct ipt_entry_match *m, unsigned int *nfcache);

	/* Function which parses command options; returns true if it
           ate an option */
	int (*parse)(int c, char **argv, int invert, unsigned int *flags,
		     const struct ipt_entry *entry,
		     unsigned int *nfcache,
		     struct ipt_entry_match **match);

	/* Final check; exit if not ok. */
	void (*final_check)(unsigned int flags);

	/* Prints out the match iff non-NULL: put space at end */
	void (*print)(const struct ipt_ip *ip,
		      const struct ipt_entry_match *match, int numeric);

	/* Saves the match info in parsable form to stdout. */
	void (*save)(const struct ipt_ip *ip,
		     const struct ipt_entry_match *match);

	/* Pointer to list of extra command-line options */
	const struct option *extra_opts;

	/* Ignore these men behind the curtain: */
	unsigned int option_offset;
	struct ipt_entry_match *m;
	unsigned int mflags;
#ifdef NO_SHARED_LIBS
	unsigned int loaded; /* simulate loading so options are merged properly */
#endif
};

struct iptables_target
{
	struct iptables_target *next;

	ipt_chainlabel name;

#ifndef IPTABLES_1_2
	/* Revision of target (0 by default). */
	u_int8_t revision;
#endif

	const char *version;

	/* Size of target data. */
	size_t size;

	/* Size of target data relevent for userspace comparison purposes */
	size_t userspacesize;

	/* Function which prints out usage message. */
	void (*help)(void);

	/* Initialize the target. */
	void (*init)(struct ipt_entry_target *t, unsigned int *nfcache);

	/* Function which parses command options; returns true if it
           ate an option */
	int (*parse)(int c, char **argv, int invert, unsigned int *flags,
		     const struct ipt_entry *entry,
		     struct ipt_entry_target **target);

	/* Final check; exit if not ok. */
	void (*final_check)(unsigned int flags);

	/* Prints out the target iff non-NULL: put space at end */
	void (*print)(const struct ipt_ip *ip,
		      const struct ipt_entry_target *target, int numeric);

	/* Saves the targinfo in parsable form to stdout. */
	void (*save)(const struct ipt_ip *ip,
		     const struct ipt_entry_target *target);

	/* Pointer to list of extra command-line options */
	struct option *extra_opts;

	/* Ignore these men behind the curtain: */
	unsigned int option_offset;
	struct ipt_entry_target *t;
	unsigned int tflags;
	unsigned int used;
#ifdef NO_SHARED_LIBS
	unsigned int loaded; /* simulate loading so options are merged properly */
#endif
};


/* Your shared library should call one of these. */
void register_match(struct iptables_match *me);
void register_target(struct iptables_target *me);

struct in_addr *dotted_to_addr(const char *dotted);
char *addr_to_dotted(const struct in_addr *addrp);
char *addr_to_anyname(const struct in_addr *addr);
char *mask_to_dotted(const struct in_addr *mask);

void parse_hostnetworkmask(const char *name, struct in_addr **addrpp,
			   struct in_addr *maskp, unsigned int *naddrs);
u_int16_t parse_protocol(const char *s);
void parse_interface(const char *arg, char *vianame, unsigned char *mask);

/* Keeping track of external matches and targets: linked lists.  */
struct iptables_match *iptables_matches = NULL;
struct iptables_target *iptables_targets = NULL;

enum ipt_tryload {
	DONT_LOAD,
	TRY_LOAD,
	LOAD_MUST_SUCCEED
};

struct iptables_target *find_target(const char *name, enum ipt_tryload);
struct iptables_match *find_match(const char *name, enum ipt_tryload,
				  struct iptables_rule_match **match);

/* kernel revision handling */
int kernel_version;
void get_kernel_version(void);
#define LINUX_VERSION(x,y,z)    (0x10000*(x) + 0x100*(y) + z)
#define LINUX_VERSION_MAJOR(x)  (((x)>>16) & 0xFF)
#define LINUX_VERSION_MINOR(x)  (((x)>> 8) & 0xFF)
#define LINUX_VERSION_PATCH(x)  ( (x)      & 0xFF)


/*
 * iptables_common.h declarations
 */

enum exittype {
	OTHER_PROBLEM = 1,
	PARAMETER_PROBLEM,
	VERSION_PROBLEM
};

/* this is a special 64bit data type that is 8-byte aligned */
#define aligned_u64 unsigned long long __attribute__((aligned(8)))

void exit_printhelp(struct iptables_rule_match 
		    *matches) __attribute__((noreturn));
void exit_tryhelp(int) __attribute__((noreturn));
int check_inverse(const char option[], int *invert, int *optind, int argc);
int string_to_number(const char *, 
		     unsigned int, 
		     unsigned int,
		     unsigned int *);
extern int string_to_number_l(const char *, 
			      unsigned long int, 
			      unsigned long int,
			      unsigned long *);
extern int string_to_number_ll(const char *, 
			       unsigned long long int, 
			       unsigned long long int,
			       unsigned long long *);
int iptables_insmod(const char *modname, const char *modprobe);
void exit_error(enum exittype, char *, ...)__attribute__((noreturn,
							  format(printf,2,3)));

const char *program_name = "nf-hipac"; 
const char *program_version = "0.9.1";
char *lib_dir;

#ifdef NO_SHARED_LIBS
# ifdef _INIT
#  define _init _INIT
# endif
  extern void init_extensions(void);
#endif



/*
 * iptables_common.h implementations (only necessary functions)
 */

void
exit_printhelp(struct iptables_rule_match *matches)
{
	struct iptables_rule_match *matchp = NULL;
	struct iptables_target *t = NULL;
	
	printf("\n%s v%s\n\n"
"Usage: %s -[AD] chain rule-specification [options]\n"
"       %s -[RI] chain rulenum rule-specification [options]\n"
"       %s -D chain rulenum [options]\n"
"       %s -[LF] [chain] [options]\n"
"       %s -[NX] chain\n"
"       %s -E old-chain-name new-chain-name\n"
"       %s -P chain target [options]\n"
"       %s -h (print this help information)\n\n",
	       program_name, program_version, program_name, program_name,
	       program_name, program_name, program_name, program_name,
	       program_name, program_name);

	printf(
"Commands:\n"
"Either long or short options are allowed.\n"
"  --append  -A chain		Append to chain\n"
"  --delete  -D chain		Delete matching rule from chain\n"
"  --delete  -D chain rulenum\n"
"				Delete rule rulenum (1 = first) from chain\n"
"  --insert  -I chain [rulenum]\n"
"				Insert in chain as rulenum (default 1 = first)\n"
"  --replace -R chain rulenum\n"
"				Replace rule rulenum (1 = first) in chain\n"
"  --list    -L [chain]		List the rules in a chain or all chains\n"
"  --flush   -F [chain]		Delete all rules in chain or all chains\n"
"  --new     -N chain		Create a new user-defined chain\n"
"  --delete-chain\n"
"            -X [chain]		Delete a user-defined chain\n"
"  --policy  -P chain target\n"
"				Change policy on chain to target\n"
"  --rename-chain\n"
"            -E old-chain new-chain\n"
"				Change chain name, (moving any references)\n"
"----------\n"
"Options:\n"
"  --proto	-p [!] proto	protocol: by number or name, eg. `tcp'\n"
"  --source	-s [!] address[/mask] or\n"
"                       address[:address]\n"
"				source(s) specification\n"
"  --destination -d [!] address[/mask] or\n"
"                       address[:address]\n"
"				destination(s) specification\n"
"  --in-interface -i [!] devname\n"
"				network interface name\n"
"  --jump	-j target\n"
"				target for rule \n"
"  --numeric	-n		numeric output of addresses and ports\n"
"  --out-interface -o [!] devname\n"
"				network interface name\n"
"  --verbose	-v		verbose mode\n"
"  --line-numbers		print line numbers when listing\n"
"  [!] --fragment -f             match second or further fragments only\n"
"  --version	-V		print package version.\n"
"----------\n"
"TCP options:\n"
" --syn				match when only SYN flag set\n"
" --not-syn			match when not only SYN flag set\n"
" --source-port [!] port[:port]\n"
" --sport ...\n"
"				match source port(s)\n"
" --destination-port [!] port[:port]\n"
" --dport ...\n"
"				match destination port(s)\n"
"----------\n"
"UDP options:\n"
" --source-port [!] port[:port]\n"
" --sport ...\n"
"				match source port(s)\n"
" --destination-port [!] port[:port]\n"
" --dport ...\n"
"				match destination port(s)\n"
"----------\n"
"ICMP options:\n"
" --icmp-type typename	        match icmp type\n"
"				(or numeric type or type/code)\n"
"				see also: nf-hipac -h icmp\n"	
"----------\n"
"STATE options:\n"
" --state [!] [INVALID|ESTABLISHED|NEW|RELATED|UNTRACKED|ESTABLISHED,RELATED]\n"
"				State to match\n"
"----------\n"
"TTL options:\n"
" --ttl   value[:value]		match time to live value(s)\n"
"----------\n");


	/* Print out any special helps. A user might like to be able
	   to add a --help to the commandline, and see expected
	   results. So we call help for all matches & targets */
	
	if (iptables_targets) {
		printf("\n-----------------\n"
		         "Iptables targets:\n"
		         "-----------------\n");
		for (t = iptables_targets; t; t = t->next) {
			if (t->used) {
				printf("\n");
				t->help();
			}
		}
	}

	if (matches) {
		printf("\n-----------------\n"
		         "Iptables matches:\n"
			 "-----------------\n");
		for (matchp = matches; matchp; matchp = matchp->next) {
			printf("\n");
			matchp->match->help();
		}
	}

	exit(0);
}


int
check_inverse(const char option[], int *invert, int *optind, int argc)
{
	if (option && strcmp(option, "!") == 0) {
		if (*invert)
			exit_error(PARAMETER_PROBLEM,
				   "Multiple `!' flags not allowed");
		*invert = TRUE;
		if (optind) {
			*optind = *optind+1;
			if (argc && *optind > argc)
				exit_error(PARAMETER_PROBLEM,
					   "no argument following `!'");
		}

		return TRUE;
	}
	return FALSE;
}


int
string_to_number_ll(const char *s, unsigned long long min,
		    unsigned long long max, unsigned long long *ret)
{
	unsigned long long number;
	char *end;
	
	/* Handle hex, octal, etc. */
	errno = 0;
	number = strtoull(s, &end, 0);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (errno != ERANGE && min <= number &&
		    (!max || number <= max)) {
			*ret = number;
			return 0;
		}
	}
	return -1;
}

int
string_to_number_l(const char *s, unsigned long min, unsigned long max,
		   unsigned long *ret)
{
	int result;
	unsigned long long number;

	result = string_to_number_ll(s, min, max, &number);
	*ret = (unsigned long)number;

	return result;
}

int
string_to_number(const char *s, unsigned int min, unsigned int max,
		 unsigned int *ret)
{
	int result;
	unsigned long number;

	result = string_to_number_l(s, min, max, &number);
	*ret = (unsigned int)number;

	return result;
}

static char *get_modprobe(void)
{
	int procfile;
	char *ret;

#define PROCFILE_BUFSIZ 1024
	procfile = open(PROC_SYS_MODPROBE, O_RDONLY);
	if (procfile < 0)
		return NULL;

	ret = (char *) malloc(PROCFILE_BUFSIZ);
	if (ret) {
		memset(ret, 0, PROCFILE_BUFSIZ);
		switch (read(procfile, ret, PROCFILE_BUFSIZ)) {
		case -1: goto fail;
		case PROCFILE_BUFSIZ: goto fail; /* Partial read.  Wierd */
		}
		if (ret[strlen(ret)-1]=='\n') 
			ret[strlen(ret)-1]=0;
		close(procfile);
		return ret;
	}
 fail:
	free(ret);
	close(procfile);
	return NULL;
}

int iptables_insmod(const char *modname, const char *modprobe)
{
	char *buf = NULL;
	char *argv[3];
	int status;

	/* If they don't explicitly set it, read out of kernel */
	if (!modprobe) {
		buf = get_modprobe();
		if (!buf)
			return -1;
		modprobe = buf;
	}

	switch (fork()) {
	case 0:
		argv[0] = (char *)modprobe;
		argv[1] = (char *)modname;
		argv[2] = NULL;
		execv(argv[0], argv);

		/* not usually reached */
		exit(1);
	case -1:
		return -1;

	default: /* parent */
		wait(&status);
	}

	free(buf); 
	if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
		return 0;
	return -1;
}

void clear_rule_matches(struct iptables_rule_match **matches)
{
	struct iptables_rule_match *matchp, *tmp;

	for (matchp = *matches; matchp;) {
		tmp = matchp->next;
		free(matchp);
		matchp = tmp;
	}

	*matches = NULL;
}


#ifndef IPTABLES_1_2
static void set_revision(char *name, u_int8_t revision)
{
	/* Old kernel sources don't have ".revision" field,
	   but we stole a byte from name. */
	name[IPT_FUNCTION_MAXNAMELEN - 2] = '\0';
	name[IPT_FUNCTION_MAXNAMELEN - 1] = revision;
}
#endif


static void free_opts(int reset_offset)
{
	if (opts != original_opts) {
		free(opts);
		opts = original_opts;
		if (reset_offset)
			global_option_offset = 0;
	}
}


void
exit_error(enum exittype status, char *msg, ...)
{
	va_list args;
	va_start(args, msg);
	fprintf(stderr, "%s v%s: ", program_name, program_version);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
	if (status == PARAMETER_PROBLEM)
		exit_tryhelp(status);
	if (status == VERSION_PROBLEM)
		fprintf(stderr,
			"Perhaps nf-hipac or your "
			"kernel needs to be upgraded.\n");
	free_opts(1);
	exit(status);
}


void
exit_tryhelp(int status)
{
	if (line != -1)
		fprintf(stderr, "Error occurred at line: %d\n", line);
	fprintf(stderr, "Try `%s -h' or '%s --help' for more information.\n",
			program_name, program_name );
	free_opts(1);
	exit(status);
}


/*
 * iptables.h implementations (only necessary functions)
 */

#ifndef IPTABLES_1_2
static int compatible_revision(const char *name, u_int8_t revision, int opt)
{
	struct ipt_get_revision rev;
	socklen_t s = sizeof(rev);
	int max_rev, sockfd;

	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0) {
		fprintf(stderr, "Could not open socket to kernel: %s\n",
			strerror(errno));
		exit(1);
	}

	strcpy(rev.name, name);
	rev.revision = revision;

	max_rev = getsockopt(sockfd, IPPROTO_IP, opt, &rev, &s);
	if (max_rev < 0) {
		/* Definitely don't support this? */
		if (errno == EPROTONOSUPPORT) {
			close(sockfd);
			return 0;
		} else if (errno == ENOPROTOOPT) {
			close(sockfd);
			/* Assume only revision 0 support (old kernel) */
			return (revision == 0);
		} else {
			fprintf(stderr, "getsockopt failed strangely: %s\n",
				strerror(errno));
			exit(1);
		}
	}
	close(sockfd);
	return 1;
}

static int compatible_match_revision(const char *name, u_int8_t revision)
{
	return compatible_revision(name, revision, IPT_SO_GET_REVISION_MATCH);
}

static int compatible_target_revision(const char *name, u_int8_t revision)
{
	return compatible_revision(name, revision, IPT_SO_GET_REVISION_TARGET);
}
#endif

void
register_match(struct iptables_match *me)
{
	struct iptables_match **i;

	/*if (strcmp(me->version, program_version) != 0) {
	 *fprintf(stderr, "%s: match `%s' v%s (I'm v%s).\n",
	 *	program_name, me->name, me->version, program_version);
	 *exit(1);
	 *} */

#ifdef IPTABLES_1_2
	if (find_match(me->name, DONT_LOAD, NULL)) {
		fprintf(stderr, "%s: match `%s' already registered.\n", 	
			program_name, me->name); 	                        
		exit(1); 	                
	}
#else
	struct iptables_match *old;
	
	/* Revision field stole a char from name. */
	if (strlen(me->name) >= IPT_FUNCTION_MAXNAMELEN-1) {
		fprintf(stderr, "%s: target `%s' has invalid name\n",
			program_name, me->name);
		exit(1);
	}

	old = find_match(me->name, DONT_LOAD, NULL);
	if (old) {
		if (old->revision == me->revision) {
			fprintf(stderr,
				"%s: match `%s' already registered.\n",
				program_name, me->name);
			exit(1);
		}

		/* Now we have two (or more) options, check compatibility. */
		if (compatible_match_revision(old->name, old->revision)
		    && old->revision > me->revision)
			return;

		/* Replace if compatible. */
		if (!compatible_match_revision(me->name, me->revision))
			return;

		/* Delete old one. */
		for (i = &iptables_matches; *i!=old; i = &(*i)->next);
		*i = old->next;
	}
#endif

	if (me->size != IPT_ALIGN(me->size)) {
		fprintf(stderr, "%s: match `%s' has invalid size %u.\n",
			program_name, me->name, (unsigned int) me->size);
		exit(1);
	}

	/* Append to list. */
	for (i = &iptables_matches; *i; i = &(*i)->next);
	me->next = NULL;
	*i = me;

	me->m = NULL;
	me->mflags = 0;
}

void
register_target(struct iptables_target *me)
{
	/*if (strcmp(me->version, program_version) != 0) {
	 *fprintf(stderr, "%s: target `%s' v%s (I'm v%s).\n",
	 *	program_name, me->name, me->version, program_version);
	 *exit(1);
	 *} */
	
#ifdef IPTABLES_1_2
	if (find_target(me->name, DONT_LOAD)) { 	
		fprintf(stderr, "%s: target `%s' already registered.\n",
			program_name, me->name);
		exit(1);
	}
#else
	struct iptables_target *old;
	
        /* Revision field stole a char from name. */
	if (strlen(me->name) >= IPT_FUNCTION_MAXNAMELEN-1) {
		fprintf(stderr, "%s: target `%s' has invalid name\n",
			program_name, me->name);
		exit(1);
	}

	old = find_target(me->name, DONT_LOAD);
	if (old) {
		struct iptables_target **i;

		if (old->revision == me->revision) {
			fprintf(stderr,
				"%s: target `%s' already registered.\n",
				program_name, me->name);
			exit(1);
		}

		/* Now we have two (or more) options, check compatibility. */
		if (compatible_target_revision(old->name, old->revision)
		    && old->revision > me->revision)
			return;

		/* Replace if compatible. */
		if (!compatible_target_revision(me->name, me->revision))
			return;

		/* Delete old one. */
		for (i = &iptables_targets; *i!=old; i = &(*i)->next);
		*i = old->next;
	}
#endif

	if (me->size != IPT_ALIGN(me->size)) {
		fprintf(stderr, "%s: target `%s' has invalid size %u.\n",
			program_name, me->name, (unsigned int) me->size);
		exit(1);
	}

	/* Prepend to list. */
	me->next = iptables_targets;
	iptables_targets = me;
	me->t = NULL;
	me->tflags = 0;
}

struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	unsigned int onebyte;
	int i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	buf[sizeof(buf) - 1] = '\0';
	addrp = (unsigned char *) &(addr.s_addr);
	
	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;

		*q = '\0';
		if (string_to_number(p, 0, 255, &onebyte) == -1)
			return (struct in_addr *) NULL;

		addrp[i] = (unsigned char) onebyte;
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if (string_to_number(p, 0, 255, &onebyte) == -1)
		return (struct in_addr *) NULL;

	addrp[3] = (unsigned char) onebyte;

	return &addr;
}

char *
addr_to_dotted(const struct in_addr *addrp)
{
	static char buf[20];
	const unsigned char *bytep;

	bytep = (const unsigned char *) &(addrp->s_addr);
	sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
	return buf;
}

static char *
addr_to_host(const struct in_addr *addr)
{
	struct hostent *host;

	if ((host = gethostbyaddr((char *) addr,
				  sizeof(struct in_addr), AF_INET)) != NULL)
		return (char *) host->h_name;

	return (char *) NULL;
}

static char *
addr_to_network(const struct in_addr *addr)
{
	struct netent *net;

	if ((net = getnetbyaddr((long) ntohl(addr->s_addr), AF_INET)) != NULL)
		return (char *) net->n_name;

	return (char *) NULL;
}

char *
addr_to_anyname(const struct in_addr *addr)
{
	char *name;

	if ((name = addr_to_host(addr)) != NULL ||
	    (name = addr_to_network(addr)) != NULL)
		return name;

	return addr_to_dotted(addr);
}

char *
mask_to_dotted(const struct in_addr *mask)
{
	int i;
	static char buf[20];
	u_int32_t maskaddr, bits;

	maskaddr = ntohl(mask->s_addr);

	if (maskaddr == 0xFFFFFFFFL)
		/* we don't want to see "/32" */
		return "";

	i = 32;
	bits = 0xFFFFFFFEL;
	while (--i >= 0 && maskaddr != bits)
		bits <<= 1;
	if (i >= 0)
		sprintf(buf, "/%d", i);
	else
		/* mask was not a decent combination of 1's and 0's */
		sprintf(buf, "/%s", addr_to_dotted(mask));

	return buf;
}

static struct in_addr *
network_to_addr(const char *name)
{
	struct netent *net;
	static struct in_addr addr;

	if ((net = getnetbyname(name)) != NULL) {
		if (net->n_addrtype != AF_INET)
			return (struct in_addr *) NULL;
		addr.s_addr = htonl((unsigned long) net->n_net);
		return &addr;
	}

	return (struct in_addr *) NULL;
}

static void *
fw_malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL) {
		perror("nf-hipac: malloc failed");
		exit(1);
	}
	return p;
}

static void *
fw_calloc(size_t count, size_t size)
{
	void *p;

	if ((p = calloc(count, size)) == NULL) {
		perror("nf-hipac: calloc failed");
		exit(1);
	}
	return p;
}

static void
inaddrcpy(struct in_addr *dst, struct in_addr *src)
{
	/* memcpy(dst, src, sizeof(struct in_addr)); */
	dst->s_addr = src->s_addr;
}

static struct in_addr *
host_to_addr(const char *name, unsigned int *naddr)
{
	struct hostent *host;
	struct in_addr *addr;
	unsigned int i;

	*naddr = 0;
	if ((host = gethostbyname(name)) != NULL) {
		if (host->h_addrtype != AF_INET ||
		    host->h_length != sizeof(struct in_addr))
			return (struct in_addr *) NULL;

		while (host->h_addr_list[*naddr] != (char *) NULL)
			(*naddr)++;
		addr = fw_calloc(*naddr, sizeof(struct in_addr) * *naddr);
                       //FIXME: why * *naddr, should be enough without it!
		for (i = 0; i < *naddr; i++)
			inaddrcpy(&(addr[i]),
				  (struct in_addr *) host->h_addr_list[i]);
		return addr;
	}

	return (struct in_addr *) NULL;
}

static struct in_addr *
parse_hostnetwork(const char *name, unsigned int *naddrs)
{
	struct in_addr *addrp, *addrptmp;

	if ((addrptmp = dotted_to_addr(name)) != NULL ||
	    (addrptmp = network_to_addr(name)) != NULL) {
		addrp = fw_malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	}
	if ((addrp = host_to_addr(name, naddrs)) != NULL)
		return addrp;

	exit_error(PARAMETER_PROBLEM, "host/network `%s' not found", name);
}

static struct in_addr *
parse_mask(char *mask)
{
	static struct in_addr maskaddr;
	struct in_addr *addrp;
	unsigned int bits;

	if (mask == NULL) {
		/* no mask at all defaults to 32 bits */
		maskaddr.s_addr = 0xFFFFFFFF;
		return &maskaddr;
	}
	if ((addrp = dotted_to_addr(mask)) != NULL)
		/* dotted_to_addr already returns a network byte order addr */
		return addrp;
	if (string_to_number(mask, 0, 32, &bits) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "invalid mask `%s' specified", mask);
	if (bits != 0) {
		maskaddr.s_addr = htonl(0xFFFFFFFF << (32 - bits));
		return &maskaddr;
	}

	maskaddr.s_addr = 0L;
	return &maskaddr;
}

void
parse_hostnetworkmask(const char *name, struct in_addr **addrpp,
		      struct in_addr *maskp, unsigned int *naddrs)
{
	struct in_addr *addrp;
	char buf[256];
	char *p;
	int i, j, k, n;

	strncpy(buf, name, sizeof(buf) - 1);
	buf[sizeof(buf) - 1] = '\0';
	if ((p = strrchr(buf, '/')) != NULL) {
		*p = '\0';
		addrp = parse_mask(p + 1);
	} else
		addrp = parse_mask(NULL);
	inaddrcpy(maskp, addrp);

	/* if a null mask is given, the name is ignored, like in "any/0" */
	if (maskp->s_addr == 0L)
		strcpy(buf, "0.0.0.0");

	addrp = *addrpp = parse_hostnetwork(buf, naddrs);
	n = *naddrs;
	for (i = 0, j = 0; i < n; i++) {
		addrp[j++].s_addr &= maskp->s_addr;
		for (k = 0; k < j - 1; k++) {
			if (addrp[k].s_addr == addrp[j - 1].s_addr) {
				(*naddrs)--;
				j--;
				break;
			}
		}
	}
}

u_int16_t
parse_protocol(const char *s)
{
	unsigned int proto;

	if (string_to_number(s, 0, 255, &proto) == -1) {
		struct protoent *pent;

		if ((pent = getprotobyname(s)))
			proto = pent->p_proto;
		else {
			unsigned int i;
			for (i = 0;
			     i < sizeof(chain_protos)/sizeof(struct pprot);
			     i++) {
				if (strcmp(s, chain_protos[i].name) == 0) {
					proto = chain_protos[i].num;
					break;
				}
			}
			if (i == sizeof(chain_protos)/sizeof(struct pprot))
				exit_error(PARAMETER_PROBLEM,
					   "unknown protocol `%s' specified",
					   s);
		}
	}

	return (u_int16_t)proto;
}

void parse_interface(const char *arg, char *vianame, unsigned char *mask)
{
	int vialen = strlen(arg);
	unsigned int i;

	memset(mask, 0, IFNAMSIZ);
	memset(vianame, 0, IFNAMSIZ);

	if (vialen + 1 > IFNAMSIZ)
		exit_error(PARAMETER_PROBLEM,
			   "interface name `%s' must be shorter than IFNAMSIZ"
			   " (%i)", arg, IFNAMSIZ-1);

	strcpy(vianame, arg);
	if ((vialen == 0) || (vialen == 1 && vianame[0] == '+'))
		memset(mask, 0, IFNAMSIZ);
	else if (vianame[vialen - 1] == '+') {
		memset(mask, 0xFF, vialen - 1);
		memset(mask + vialen - 1, 0, IFNAMSIZ - vialen + 1);
		/* Don't remove `+' here! -HW */
	} else {
		/* Include nul-terminator in match */
		memset(mask, 0xFF, vialen + 1);
		memset(mask + vialen + 1, 0, IFNAMSIZ - vialen - 1);
		for (i = 0; vianame[i]; i++) {
			if (!isalnum(vianame[i]) 
			    && vianame[i] != '_' 
			    && vianame[i] != '.') {
				printf("Warning: wierd character in interface"
				       " `%s' (No aliases, :, ! or *).\n",
				       vianame);
				break;
			}
		}
	}
}

struct iptables_target *
find_target(const char *name, enum ipt_tryload tryload)
{
	struct iptables_target *ptr;

	/* Standard target? */
	if (strcmp(name, "") == 0
	    || strcmp(name, IPTC_LABEL_ACCEPT) == 0
	    || strcmp(name, IPTC_LABEL_DROP) == 0
	    // not supported by nf-hipac
	    //  || strcmp(name, IPTC_LABEL_QUEUE) == 0 
	    || strcmp(name, IPTC_LABEL_RETURN) == 0)
		name = "standard";

	for (ptr = iptables_targets; ptr; ptr = ptr->next) {
		if (strcmp(name, ptr->name) == 0)
			break;
	}

#ifndef NO_SHARED_LIBS
	if (!ptr && tryload != DONT_LOAD) {
		char path[strlen(lib_dir) + sizeof("/libipt_.so")
			 + strlen(name)];
		sprintf(path, "%s/libipt_%s.so", lib_dir, name);
		if (dlopen(path, RTLD_NOW)) {
			/* Found library.  If it didn't register itself,
			   maybe they specified match as a target. */
			ptr = find_target(name, DONT_LOAD);
			if (!ptr)
				exit_error(PARAMETER_PROBLEM,
					   "Couldn't load target `%s'\n",
					   name);
		} else if (tryload == LOAD_MUST_SUCCEED)
			exit_error(PARAMETER_PROBLEM,
				   "Couldn't load target `%s':%s\n",
				   name, dlerror());
	}
#else
	if (ptr && !ptr->loaded) {
		if (tryload != DONT_LOAD)
			ptr->loaded = 1;
		else
			ptr = NULL;
	}
	if(!ptr && (tryload == LOAD_MUST_SUCCEED)) {
		exit_error(PARAMETER_PROBLEM,
			   "Couldn't find target `%s'\n", name);
	}
#endif

	if (ptr)
		ptr->used = 1;

	return ptr;
}

struct iptables_match *
find_match(const char *name, enum ipt_tryload tryload,
	   struct iptables_rule_match **matches)
{
	struct iptables_match *ptr;

	for (ptr = iptables_matches; ptr; ptr = ptr->next) {
		if (strcmp(name, ptr->name) == 0)
			break;
	}

#ifndef NO_SHARED_LIBS
	if (!ptr && tryload != DONT_LOAD) {
		char path[strlen(lib_dir) + sizeof("/libipt_.so")
			 + strlen(name)];
		sprintf(path, "%s/libipt_%s.so", lib_dir, name);
		if (dlopen(path, RTLD_NOW)) {
			/* Found library.  If it didn't register itself,
			   maybe they specified target as match. */
			ptr = find_match(name, DONT_LOAD, NULL);

			if (!ptr)
				exit_error(PARAMETER_PROBLEM,
					   "Couldn't load match `%s'\n",
					   name);
		} else if (tryload == LOAD_MUST_SUCCEED)
			exit_error(PARAMETER_PROBLEM,
				   "Couldn't load match `%s':%s\n",
				   name, dlerror());
	}
#else
	if (ptr && !ptr->loaded) {
		if (tryload != DONT_LOAD)
			ptr->loaded = 1;
		else
			ptr = NULL;
	}
	if(!ptr && (tryload == LOAD_MUST_SUCCEED)) {
		exit_error(PARAMETER_PROBLEM,
			   "Couldn't find match `%s'\n", name);
	}
#endif

	if (ptr && matches) {
		struct iptables_rule_match **i;
		struct iptables_rule_match *newentry;

		newentry = fw_malloc(sizeof(struct iptables_rule_match));

		for (i = matches; *i; i = &(*i)->next);
		newentry->match = ptr;
		newentry->next = NULL;
		*i = newentry;
	}

	return ptr;
}

void
get_kernel_version(void)
{
	static struct utsname uts;
	int x = 0, y = 0, z = 0;
	
	if (uname(&uts) == -1) {
		fprintf(stderr, "Unable to retrieve kernel version.\n");
		free_opts(1);
		exit(1);
	}
	
	sscanf(uts.release, "%d.%d.%d", &x, &y, &z);
	kernel_version = LINUX_VERSION(x, y, z);
}


/*
 * ========== End of iptables matches/targets compatibility code ==========
 */




static char
cmd2char(int option)
{
	const char *ptr;
	for (ptr = cmdflags; option > 1; option >>= 1, ptr++);
	return *ptr;
}


static void
add_command(struct nfhp_cmd *message, int newcmd, int invert)
{		
	if (invert)
		exit_error(PARAMETER_PROBLEM, "unexpected ! flag");
	if (message->cmd)
		exit_error(PARAMETER_PROBLEM, "Can't use -%c with -%c\n",
			   cmd2char(newcmd), cmd2char(message->cmd));
	message->cmd = newcmd;	
}

static unsigned int
dimid_to_option(unsigned int dimid)
{
	return (1 << (dimid + NUMBER_OF_OPT));
}

static void
set_option(unsigned int *options, unsigned int std_opt, unsigned int dimid)
{
	unsigned int option;
	if (std_opt) 
		option = std_opt;
	else 	option = dimid_to_option(dimid);
	if (*options & option)
		exit_error(PARAMETER_PROBLEM,
			   "multiple flags of the same "
			   "type not allowed");
	*options |= option;
}



void
parse_rulenumber(const char *str, unsigned int *rulenum)
{
	if (string_to_number(str, 0, INT_MAX, rulenum) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid rule number `%s'", str);
}




static struct option *
merge_options(struct option *oldopts, const struct option *newopts,
	      unsigned int *option_offset)
{
	unsigned int num_old, num_new, i;
	struct option *merge;

	for (num_old = 0; oldopts[num_old].name; num_old++);
	for (num_new = 0; newopts[num_new].name; num_new++);
	
	global_option_offset += OPTION_OFFSET;
	*option_offset = global_option_offset;

	merge = malloc(sizeof(struct option) * (num_new + num_old + 1));
	memcpy(merge, oldopts, num_old * sizeof(struct option));
	free_opts(0);
	for (i = 0; i < num_new; i++) {
		merge[num_old + i] = newopts[i];
		merge[num_old + i].val += *option_offset;
	}
	memset(merge + num_old + num_new, 0, sizeof(struct option));
	return merge;
}

static void
parse_target(const char *arg, char *target)
{
	const char *ptr;
	memset(target, 0, HIPAC_CHAIN_NAME_MAX_LEN);

	if (strlen(arg) < 1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid target name (too short)");
	
	if (strlen(arg) + 1 > HIPAC_CHAIN_NAME_MAX_LEN)
		exit_error(PARAMETER_PROBLEM,
			   "target name `%s' must be shorter than "
			   "HIPAC_CHAIN_NAME_MAX_LEN (%i)", 
			   arg, HIPAC_CHAIN_NAME_MAX_LEN);	
	for (ptr = arg; *ptr; ptr++)
		if (isspace(*ptr))
			exit_error(PARAMETER_PROBLEM,
				   "Invalid target name `%s'", arg);
	strcpy(target, arg);
}

void
parse_policy(const char *targetname, u_int8_t *policy)
{
	if (strlen(targetname) < 1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid target name (too short)");
	if (strcasecmp(targetname, IPTC_LABEL_DROP) == 0){
		*policy = TARGET_DROP;
		return;
	} else if (strcasecmp(targetname, IPTC_LABEL_ACCEPT) == 0){
		*policy = TARGET_ACCEPT;
		return;
	} 
	
	exit_error(PARAMETER_PROBLEM,
		   "Invalid target name `%s'", targetname);
}

void
parse_native_protocol(const char *s, struct dim_match *rule)
{
	char buf[256];
	char *p;

	strncpy(buf, s, sizeof(buf) - 1);
	buf[sizeof(buf) - 1] = '\0';
	if ((p = strrchr(buf, ':')) != NULL) {
		*p = '\0'; 
		p++;
		rule[DIMID_PROTO].left = buf[0] ? parse_protocol(buf) : 0;
		if (strcasecmp(buf, p) == 0){
			rule[DIMID_PROTO].right = rule[DIMID_PROTO].left;
		} else {
			rule[DIMID_PROTO].right = p[0] ? parse_protocol(p) : 255;
			if (rule[DIMID_PROTO].left > rule[DIMID_PROTO].right)
				exit_error(PARAMETER_PROBLEM,
					   "Range Error: start value > "
					   "end value\n");
			if ((rule[DIMID_PROTO].left == 0) &&
			    (rule[DIMID_PROTO].right == 255)){
				return;
			}
		}
	} else {
		rule[DIMID_PROTO].left =
			rule[DIMID_PROTO].right = parse_protocol(s);
	}
	set_option(&options, 0, DIMID_PROTO);
}


static void
parse_native_interface(const char *arg, char *vianame)
{
	int vialen = strlen(arg);
	unsigned int i;

	memset(vianame, 0, IFNAMSIZ);

	if (vialen + 1 > IFNAMSIZ)
		exit_error(PARAMETER_PROBLEM,
			   "interface name `%s' must be shorter than IFNAMSIZ"
			   " (%i)", arg, IFNAMSIZ-1);

	strcpy(vianame, arg);
	if (vialen == 0)
		exit_error(PARAMETER_PROBLEM,
			   "no interface name specified");
	else if (vianame[vialen - 1] == '+') {
		exit_error(PARAMETER_PROBLEM,
			   "no + matches on interfaces supported");
	} else {
		/* Include nul-terminator in match */
		for (i = 0; vianame[i]; i++) {
			if (!isalnum(vianame[i]) 
			    && vianame[i] != '_' 
			    && vianame[i] != '.') {
				printf("Warning: wierd character in interface"
				       " `%s' (No aliases, :, ! or *).\n",
				       vianame);
				break;
			}
		}
	}
}

static void
parse_chain(const char *arg, char *vianame)
{
	int vialen = strlen(arg);

	memset(vianame, 0, HIPAC_CHAIN_NAME_MAX_LEN);

	if (vialen + 1 > HIPAC_CHAIN_NAME_MAX_LEN)
		exit_error(PARAMETER_PROBLEM,
			   "chain name `%s' must be shorter than "
			   "HIPAC_CHAIN_NAME_MAX_LEN (%i)", 
			   arg, HIPAC_CHAIN_NAME_MAX_LEN);
	
	if (vialen == 0)
		exit_error(PARAMETER_PROBLEM,
			   "no chain name specified");
	strcpy(vianame, arg);
}

static int
parse_state(const char *state, size_t strlen, unsigned int* sinfo)
{
	if (strncasecmp(state, "INVALID", strlen) == 0)
		*sinfo = NFHP_STATE_INVALID;
	else if (strncasecmp(state, "UNTRACKED", strlen) == 0)
		*sinfo = NFHP_STATE_UNTRACKED;
	else if (strncasecmp(state, "NEW", strlen) == 0)
		*sinfo = NFHP_STATE_NEW;
	else if (strncasecmp(state, "ESTABLISHED", strlen) == 0)
		*sinfo = NFHP_STATE_ESTABLISHED;
	else if (strncasecmp(state, "RELATED", strlen) == 0)
		*sinfo = NFHP_STATE_RELATED;
	else
		return 0;
	return 1;
}

static void
parse_states(const char *arg, struct dim_match *rule)

{
	unsigned int left, right;
	const char *comma;

	if ((comma = strchr(arg, ',')) != NULL) {
		if (comma == arg || !parse_state(arg, comma - arg, &left))
                        exit_error(PARAMETER_PROBLEM, "Bad state `%s'", arg);
		if (strlen(comma + 1) == 0 || 
		    !parse_state(comma + 1, strlen(comma + 1), &right))
			exit_error(PARAMETER_PROBLEM, "Bad state `%s'", arg);
		if (left == right) {
			rule[DIMID_STATE].left = 
				rule[DIMID_STATE].right = left; 
		} else if ((left == NFHP_STATE_ESTABLISHED &&
			    right == NFHP_STATE_RELATED) ||
			   (left  == NFHP_STATE_RELATED &&
			    right == NFHP_STATE_ESTABLISHED)) {
			rule[DIMID_STATE].left = NFHP_STATE_ESTABLISHED;
			rule[DIMID_STATE].right = NFHP_STATE_RELATED;
		} else {
			exit_error(PARAMETER_PROBLEM, "Bad state `%s'", arg);
		}
	} else {
		if (strlen(arg) == 0 || 
		    !parse_state(arg, strlen(arg), &left))
			exit_error(PARAMETER_PROBLEM, "Bad state `%s'", arg);
		rule[DIMID_STATE].left =
			rule[DIMID_STATE].right = left;
	}
}



static struct in_addr *
hbo_parse_hostnetwork(const char *name, unsigned int *naddrs)
{
	struct in_addr *addrp, *addrptmp;
	int i;

	if ((addrptmp = dotted_to_addr(name)) != NULL ||
	    (addrptmp = network_to_addr(name)) != NULL) {
		addrp = fw_malloc(sizeof(struct in_addr));
		addrp->s_addr = ntohl(addrptmp->s_addr);
		*naddrs = 1;
		return addrp;
	}
	if ((addrp = host_to_addr(name, naddrs)) != NULL){
		for (i = 0; i < *naddrs; i++)
			addrp[i].s_addr = ntohl(addrp[i].s_addr);
		return addrp;
	}

	exit_error(PARAMETER_PROBLEM, "host/network `%s' not found", name);
	return (struct in_addr *) NULL;
}

static struct in_addr *
hbo_parse_mask(char *mask)
{
	static struct in_addr maskaddr;
	struct in_addr *addrp;
	unsigned int bits;
	int i;
	u_int32_t f;
	
	if ((addrp = dotted_to_addr(mask)) != NULL){
		addrp->s_addr = ntohl(addrp->s_addr);
		if (addrp->s_addr == 0xFFFFFFFFL){
			return addrp;
		}
		i = 32;
		f = 0xFFFFFFFEL;
		while (--i >= 0 && addrp->s_addr != f)
			f <<= 1;
		if (i >= 0)
			return addrp;
		else
			exit_error(PARAMETER_PROBLEM,
				   "invalid mask `%s' specified", mask);
	}
	if (string_to_number(mask, 0, 32, &bits) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "invalid mask `%s' specified", mask);
	if (bits != 0) {
		maskaddr.s_addr = (0xFFFFFFFF << (32 - bits));
		return &maskaddr;
	}

	maskaddr.s_addr = 0L;
	return &maskaddr;
}

void
hbo_parse_hostnetworkmask(const char *name, const int dimid,
			  struct dim_match *rule, struct in_addr **addrpp,
			  unsigned int *naddrs)
{
	struct in_addr *addrp;
	struct in_addr addr, mask;
	char buf[256];
	char *p;

	
	strncpy(buf, name, sizeof(buf) - 1);
	buf[sizeof(buf) - 1] = '\0';
	if ((p = strrchr(buf, '/')) != NULL) {
		*p = '\0';
		addrp = hbo_parse_mask(p + 1);
		mask = *addrp;
		if (mask.s_addr == 0L)
			strcpy(buf, "0.0.0.0");
		addrp = hbo_parse_hostnetwork(buf, naddrs);
		addr = *addrp;
		free(addrp);
		*naddrs = 1;
		rule[dimid].left = addr.s_addr & mask.s_addr;
		rule[dimid].right = addr.s_addr | (~mask.s_addr);
		if ((rule[dimid].left == 0) &&
		    (rule[dimid].right == 0xFFFFFFFF)){
			*naddrs = 0;
			return;
		}
	} else if ((p = strrchr(buf, ':')) != NULL) {
		*p = '\0';
		*naddrs = 1;
		addrp = hbo_parse_hostnetwork(buf, naddrs);
		addr = *addrp;
		free(addrp);
		if (strcasecmp(buf, p + 1) == 0){
			mask = addr;
		} else {
			addrp = hbo_parse_hostnetwork(p + 1, naddrs);
			mask = *addrp;
			free(addrp);
			if (addr.s_addr > mask.s_addr)
				exit_error(PARAMETER_PROBLEM,
					   "Range Error: start value > "
					   "end value\n");
			if ((addr.s_addr == 0) &&
			    (mask.s_addr == 0xFFFFFFFF)){
				*naddrs = 0;
				return;
			}
		}
   		rule[dimid].left = addr.s_addr;
		rule[dimid].right = mask.s_addr;
	} else {
		addrp = hbo_parse_hostnetwork(buf, naddrs);
		rule[dimid].left = addrp->s_addr;
		rule[dimid].right = addrp->s_addr;
		if (*naddrs > 1)
			*addrpp = addrp;
		else
			free(addrp);
	}
	set_option(&options, 0, dimid);
}

static int
service_to_port(const char *name)
{
	struct servent *service;

	if ((service = getservbyname(name, NULL)) != NULL)
		return ntohs((unsigned short) service->s_port);

	return -1;
}

static int
parse_port(const char *port)
{
	unsigned int portnum;

	if (string_to_number(port, 0, 65535, &portnum) != -1 ||
	    (portnum = service_to_port(port)) != -1)
		return (int )portnum;

	exit_error(PARAMETER_PROBLEM,
		   "invalid port/service `%s' specified", port);
	return 0;
}

static void
parse_ports(const char *portstring,  const int dimid, struct dim_match *rule)
{
	char *buffer;
	char *cp;

	buffer = strdup(portstring);
	if ((cp = strchr(buffer, ':')) == NULL)
		 rule[dimid].left =
			 rule[dimid].right = parse_port(buffer);
	else {
		*cp = '\0';
		cp++;

		rule[dimid].left = buffer[0] ? parse_port(buffer) : 0;
		rule[dimid].right = cp[0] ? parse_port(cp) : 0xFFFF;

		if (rule[dimid].left > rule[dimid].right)
			exit_error(PARAMETER_PROBLEM,
				   "invalid portrange (min > max)");
	}
	free(buffer);
	if ((rule[dimid].left == 0) && 
	    (rule[dimid].right == 65535))
		return;
	set_option(&options, 0, dimid);
}

static int
parse_ttl(const char *ttlstring)
{
	unsigned int ttl;

	if (string_to_number(ttlstring, 0, 255, &ttl) != -1)
		return (int )ttl;

	exit_error(PARAMETER_PROBLEM,
		   "invalid TTL `%s' specified", ttlstring);
	return 0;
}

static void
parse_ttls(const char *ttlstring,  struct dim_match *rule)
{
	char *buffer;
	char *cp;

	buffer = strdup(ttlstring);
	if ((cp = strchr(buffer, ':')) == NULL)
		rule[DIMID_TTL].left =
                        rule[DIMID_TTL].right = parse_ttl(buffer);
	else {
		*cp = '\0';
		cp++;

		rule[DIMID_TTL].left = buffer[0] ? parse_ttl(buffer) : 0;
		rule[DIMID_TTL].right = cp[0] ? parse_ttl(cp) : 0xFF;

		if (rule[DIMID_TTL].left > rule[DIMID_TTL].right)
			exit_error(PARAMETER_PROBLEM,
				   "invalid TTL range (min > max)");
		if ((rule[DIMID_TTL].left == 0) &&
		    (rule[DIMID_TTL].right == 0xFF)){
			free(buffer);
			return;
		}
	}
	free(buffer);
	set_option(&options, 0, DIMID_TTL);
}

static void
parse_syn(const char *flag, struct dim_match *rule)
{
	if (strcmp(flag, "SYN") == 0) {
		rule[DIMID_SYN].left = 0;
		rule[DIMID_SYN].right = 0;
		return;
	} else if (strcmp(flag, "NOT-SYN") == 0){
		rule[DIMID_SYN].left = 1;
		rule[DIMID_SYN].right = 65535;
		return;
	} else
		exit_error(PARAMETER_PROBLEM,
				   "Unknown TCP flag `%s'", flag);
}

static void
parse_icmp(const char *icmptype, struct dim_match *rule)
{
	unsigned int limit = sizeof(icmp_codes)/sizeof(struct icmp_names);
	unsigned int match = limit;
	unsigned int i;

	for (i = 0; i < limit; i++) {
		if (strncasecmp(icmp_codes[i].name, icmptype, strlen(icmptype))
		    == 0) {
			if (match != limit)
				exit_error(PARAMETER_PROBLEM,
					   "Ambiguous ICMP type `%s':"
					   " `%s' or `%s'?",
					   icmptype,
					   icmp_codes[match].name,
					   icmp_codes[i].name);
			match = i;
		}
	}

	if (match != limit) {
		rule[DIMID_ICMP_TYPE].left = (icmp_codes[match].type << 8) + 
			icmp_codes[match].code_min;
		rule[DIMID_ICMP_TYPE].right = (icmp_codes[match].type << 8) +
			icmp_codes[match].code_max;
	} else {
		char *slash;
		char buffer[strlen(icmptype) + 1];
		unsigned int type, code;

		strcpy(buffer, icmptype);
		slash = strchr(buffer, '/');

		if (slash)
			*slash = '\0';

		if (string_to_number(buffer, 0, 255, &type) == -1)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid ICMP type `%s'\n", buffer);
		if (slash) {
			if (string_to_number(slash+1, 0, 255, &code) == -1)
				exit_error(PARAMETER_PROBLEM,
					   "Invalid ICMP code `%s'\n",
					   slash+1);
		 	rule[DIMID_ICMP_TYPE].left =
				rule[DIMID_ICMP_TYPE].right =
				(type << 8) + code;
		} else {
			rule[DIMID_ICMP_TYPE].left = type << 8;
			rule[DIMID_ICMP_TYPE].right = (type << 8) + 0xFF;
		}
	}
}

static void
generic_opt_check(int command, int options)
{
	int i, k;

	for (i = 0; i < NUMBER_OF_OPT; i++) {
		if (options & (1<<i)) {
			if ((1<<i) == OPT_VERBOSE
			    || (1<<i) == OPT_DEBUG
			    || (1<<i) == OPT_DEBUGN)
				continue;
			if ((1<<i) == OPT_NUMERIC
			    || (1<<i) == OPT_LINENUMBERS)
				k = 0;
			else k = 1;
			
			if (commands_v_options[command][k] == 'x')
				exit_error(PARAMETER_PROBLEM,
					   "Illegal option `-%s'"
					   " with this command",
					   optflags[i]);
		}
	}
}

static inline void
fill_hipac_matches(struct nfhp_cmd *message, struct dim_match *rule)
{
	int i;
	message->rule.r.native_mct = 0;
	for (i = 0; i < NUMBER_OF_DIM; i++){
		if (!(options & dimid_to_option(i)))
			continue;
		message->rule.m[message->rule.r.native_mct].dimid = i;
		message->rule.m[message->rule.r.native_mct].invert = 
			rule[i].invert;
		message->rule.m[message->rule.r.native_mct].left =
			rule[i].left;
		message->rule.m[message->rule.r.native_mct].right =
			rule[i].right;
		message->rule.r.native_mct++;
	}
}

static char *
port_to_service(int port)
{
	struct servent *service;

	if ((service = getservbyport(htons(port), NULL)))
		return service->s_name;

	return NULL;
}

static void
print_port(u_int16_t port, int numeric)
{
	char *service;

	if (numeric || (service = port_to_service(port)) == NULL)
		printf("%u", port);
	else
		printf("%s", service);
}

static void
print_ports(const char *name, u_int32_t min, u_int32_t max, u_int8_t invert,
	    int numeric)
{
	printf("%s", name);
	if (min == max) {
		printf(": ");
		if (invert)
			printf("!");
		print_port(min, numeric);
	} else {
		
		printf("s: ");
		if (invert)
			printf("!");
		print_port(min, numeric);
		printf(":");
		print_port(max, numeric);
	}
	printf("  ");
}

static void
print_ttls(u_int32_t min, u_int32_t max)
{
	printf("ttl");
	if (min == max) {
		printf(": ");
		printf("%u", min);
	} else {
		
		printf("s: %u:%u", min, max);
	}
	printf("  ");
}

static void 
print_icmptype(u_int8_t type, u_int8_t code_min, u_int8_t code_max,
	       int numeric)
{
	printf("icmp: ");
	if (!numeric) {
		unsigned int i;

		for (i = 0;
		     i < sizeof(icmp_codes)/sizeof(struct icmp_names);
		     i++) {
			if (icmp_codes[i].type == type
			    && icmp_codes[i].code_min == code_min
			    && icmp_codes[i].code_max == code_max)
				break;
		}

		if (i != sizeof(icmp_codes)/sizeof(struct icmp_names)) {
			printf("%s ", icmp_codes[i].name);
			return;
		}
	}

	printf("type %u ", type);
	if (code_min == 0 && code_max == 0xFF)
		printf(" ");
	else if (code_min == code_max)
		printf("code %u ", code_min);
	else
		printf("codes %u-%u ", code_min, code_max);
	printf(" ");
}

static void
print_syn(u_int32_t min, u_int32_t max)
{
	printf("flags: ");
	if (!min & !max) {
		printf("syn");
	} else if ((min == 1) & (max == 65535)){
		printf("not-syn");
	} else
	       printf("unknown");
	printf("  ");
}

static void
print_state(unsigned int left, unsigned int right,
	    u_int8_t invert)
{
	printf("state: ");
	if (invert)
		printf("!");
	if (left != right)
		printf("ESTABLISHED,RELATED");
	else if (left ==  NFHP_STATE_INVALID)
		printf("INVALID");
	else if (left == NFHP_STATE_NEW)
		printf("NEW");
	else if (left == NFHP_STATE_RELATED)
		printf("RELATED");
	else if (left == NFHP_STATE_ESTABLISHED)
		printf("ESTABLISHED");
	else if (left == NFHP_STATE_UNTRACKED)
		printf("UNTRACKED");
	printf("  ");
}

static char *
print_ip(int dimid, int numeric, struct dim_match *rule)
{
	u_int32_t left = htonl(rule[dimid].left);
	u_int32_t right = htonl(rule[dimid].right);
	u_int32_t mask;
	int i = 32;
	u_int32_t f = 0xFFFFFFFEL;
	static char buf[BUFSIZ];

	if (left == right){
		if (numeric)
			sprintf(buf, "%s", 
				addr_to_dotted((struct in_addr *) &left));
		else sprintf(buf, "%s", 
			     addr_to_anyname((struct in_addr*) &left));
		return buf;
	}
		
	mask = ~(rule[dimid].right - rule[dimid].left);
	while (--i >= 0 && mask != f)
		f <<= 1;

	if (i >= 0 && (((rule[dimid].left) & f) == rule[dimid].left)){
		if (numeric)
			sprintf(buf, "%s/%d", 
				addr_to_dotted((struct in_addr *) &left), i);
		else sprintf(buf, "%s/%d", 
			     addr_to_anyname((struct in_addr*) &left), i);
		return buf;
	} else {
		if (numeric){
			sprintf(buf, "%s:", 
				addr_to_dotted((struct in_addr *) &left));
			strcat(buf, addr_to_dotted((struct in_addr *) &right));
		}else {
			sprintf(buf, "%s:", 
				addr_to_anyname((struct in_addr*) &left));
			strcat(buf, addr_to_anyname((struct in_addr*) &right));
		}
		return buf;
	}
}

static void
print_icmptypes(void)
{
	unsigned int i;
	printf("Valid ICMP Types:");

	for (i = 0; i < sizeof(icmp_codes)/sizeof(struct icmp_names); i++) {
		if (i && icmp_codes[i].type == icmp_codes[i-1].type) {
			if (icmp_codes[i].code_min == icmp_codes[i-1].code_min
			    && (icmp_codes[i].code_max
				== icmp_codes[i-1].code_max))
				printf(" (%s)", icmp_codes[i].name);
			else
				printf("\n   %s", icmp_codes[i].name);
		}
		else
			printf("\n%s", icmp_codes[i].name);
	}
	printf("\n");
	exit(0);
}

static char *
proto_to_name(u_int8_t proto)
{
	unsigned int i;

	if (proto) {
		struct protoent *pent = getprotobynumber(proto);
		if (pent)
			return pent->p_name;
	}

	for (i = 0; i < sizeof(chain_protos)/sizeof(struct pprot); i++)
		if (chain_protos[i].num == proto)
			return chain_protos[i].name;

	return NULL;
}

static int
print_match(const struct ipt_entry_match *m,
	    const struct ipt_ip *ip,
	    int numeric)
{
	struct iptables_match *match = find_match(m->u.user.name,
						  TRY_LOAD, NULL);

	if (match) {
		if (match->print)
			match->print(ip, m, numeric);
		else
			printf("%s ", match->name);
	} else {
		if (m->u.user.name[0])
			printf("UNKNOWN match `%s' ", m->u.user.name);
	}
	/* Don't stop iterating. */
	return 0;
}

static void
print_rule(struct nfhp_list_rule* pt_rule, int is_nfhp_rule)
{
	int i;
	unsigned int rule_options = 0;
	struct dim_match rule[NUMBER_OF_DIM + 1];
	struct iptables_target *target = NULL;
	struct ipt_entry_target *t = NULL;

	for (i = 0; i < pt_rule->r.native_mct; i++){
		set_option(&rule_options, 0, pt_rule->r.first_match[i].dimid);
		rule[pt_rule->r.first_match[i].dimid].invert =
			pt_rule->r.first_match[i].invert;
		rule[pt_rule->r.first_match[i].dimid].left =
			pt_rule->r.first_match[i].left;
		rule[pt_rule->r.first_match[i].dimid].right =
			pt_rule->r.first_match[i].right;
	}

	if (options & OPT_LINENUMBERS)
		printf("%-5u ", pt_rule->r.pos);

	if (pt_rule->r.action == TARGET_ACCEPT){
		printf("%-9s ", IPTC_LABEL_ACCEPT);
	} else if (pt_rule->r.action == TARGET_DROP){
		printf("%-9s ", IPTC_LABEL_DROP);
	} else if (pt_rule->r.action == TARGET_RETURN){
		printf("%-9s ", IPTC_LABEL_RETURN);
	} else if (pt_rule->r.action == TARGET_NONE){
		printf("%-9s ", " ");
	} else if (pt_rule->r.action == TARGET_DUMMY){
		printf("%-9s ", "DUMMY");
	} else if (HAS_CHAIN_TARGET(&pt_rule->r)){
		if (is_nfhp_rule){
			printf("%-9s ", 
			       (char *) ((struct nfhp_rule *) pt_rule)->e
			       + ((struct nfhp_rule *) 
				  pt_rule)->e->target_offset);
		} else {
			printf("%-9s ", CHAIN_TARGET(&pt_rule->r));
		}
	} else if (HAS_IPT_TARGET(&pt_rule->r)){
		if (is_nfhp_rule){
			t = (void *) ((struct nfhp_rule *) pt_rule)->e
				+ ((struct nfhp_rule *) 
				   pt_rule)->e->target_offset;
		} else {
			t = IPT_TARGET(&pt_rule->r);
		}
		target = find_target(t->u.user.name,
				     TRY_LOAD);
		printf("%-9s ", t->u.user.name);
	
	} else {
		printf("%-9s ", "UNKNOWN");
	}

	if (rule_options & dimid_to_option(DIMID_PROTO)){
		char *pname = proto_to_name(rule[DIMID_PROTO].left);
		printf("%c", rule[DIMID_PROTO].invert ? '!' : ' ');
		if (rule[DIMID_PROTO].left != rule[DIMID_PROTO].right)
			printf("%-1hu:%-2hu ", rule[DIMID_PROTO].left,
			       rule[DIMID_PROTO].right);
		else if (pname)
			printf("%-4s ", pname);
		else
			printf("%-4hu ", rule[DIMID_PROTO].left);
	} else printf(" %-4s ", "all");

	if (rule_options & dimid_to_option(DIMID_FRAGMENT)){
		if (rule[DIMID_FRAGMENT].left == 0)
			printf("%-3s ", "!f");
		else printf("%-3s ", "-f");
	} else printf("%-3s ", "--");

	if (options & OPT_VERBOSE){
		if (*pt_rule->indev){
			if (rule_options & dimid_to_option(DIMID_INIFACE))
				printf("%c", rule[DIMID_INIFACE].invert ? 
				       '!' : ' ');
			printf("%-6s ", pt_rule->indev);
		} else	{
			if (options & (OPT_NUMERIC | OPT_DEBUGN))
				printf(" %-6s ", "*");
			else printf(" %-6s ", "any");
		}
		if (*pt_rule->outdev){
			if (rule_options & dimid_to_option(DIMID_OUTIFACE))
				printf("%c", rule[DIMID_OUTIFACE].invert ? 
				       '!' : ' ');
			printf("%-6s ", pt_rule->outdev);
		} else {
			if (options & (OPT_NUMERIC | OPT_DEBUGN))
				printf(" %-6s ", "*");
			else printf(" %-6s ", "any");
		}	
	}
	
	if (rule_options & dimid_to_option(DIMID_SRC_IP)){
		printf("%c", rule[DIMID_SRC_IP].invert ? '!' : ' ');
		printf("%-19s ", print_ip(DIMID_SRC_IP, 
					  options & 
					  (OPT_NUMERIC | OPT_DEBUGN), rule));
	} else {
		if (options & (OPT_NUMERIC | OPT_DEBUGN))
			printf(" %-19s ", "0.0.0.0/0");
		else 	printf(" %-19s ", "anywhere");
	}

	if (rule_options & dimid_to_option(DIMID_DEST_IP)){
		printf("%c", rule[DIMID_DEST_IP].invert ? '!' : ' ');
		printf("%-19s ", print_ip(DIMID_DEST_IP, 
					  options & 
					  (OPT_NUMERIC | OPT_DEBUGN), rule));
	} else {
		if (options & (OPT_NUMERIC | OPT_DEBUGN))
			printf(" %-19s ", "0.0.0.0/0");
		else 	printf(" %-19s ", "anywhere");
	}	
	
	if (rule_options & dimid_to_option(DIMID_SPORT))
		print_ports("spt", rule[DIMID_SPORT].left,
			    rule[DIMID_SPORT].right,
			    rule[DIMID_SPORT].invert, 
			    options & (OPT_NUMERIC | OPT_DEBUGN));

	if (rule_options & dimid_to_option(DIMID_DPORT))
		print_ports("dpt", rule[DIMID_DPORT].left,
			    rule[DIMID_DPORT].right, 
			    rule[DIMID_DPORT].invert, 
			    options & (OPT_NUMERIC | OPT_DEBUGN));

	if (rule_options & dimid_to_option(DIMID_SYN))
		print_syn(rule[DIMID_SYN].left, rule[DIMID_SYN].right);

	if (rule_options & dimid_to_option(DIMID_ICMP_TYPE))
		print_icmptype(rule[DIMID_ICMP_TYPE].left >> 8,
			       rule[DIMID_ICMP_TYPE].left & 0xFF,
			       rule[DIMID_ICMP_TYPE].right & 0xFF,
			       options & (OPT_NUMERIC | OPT_DEBUGN));
	
	if (rule_options & dimid_to_option(DIMID_STATE))
		print_state(rule[DIMID_STATE].left, rule[DIMID_STATE].right,
			    rule[DIMID_STATE].invert);
	
	if (rule_options & dimid_to_option(DIMID_TTL))
		print_ttls(rule[DIMID_TTL].left, rule[DIMID_TTL].right);
 
	if (target || HAS_IPT_MATCH(&pt_rule->r)){  
		struct ipt_ip ip;
		memset(&ip, 0, sizeof(ip));
		if (rule_options & dimid_to_option(DIMID_PROTO))
			if (!rule[DIMID_PROTO].invert)
				ip.proto = rule[DIMID_PROTO].left;
		if (HAS_IPT_MATCH(&pt_rule->r)){
			struct ipt_entry_match *match;
			void * match_end;
			
			if (is_nfhp_rule){
				match = (void *) 
					((struct nfhp_rule *) pt_rule)->e 
					+ IPT_ALIGN(sizeof(struct ipt_entry));
				if (target){
					match_end = (void *) 
						((struct nfhp_rule *) 
						 pt_rule)->e 
						+ ((struct nfhp_rule *) 
						   pt_rule)->e->target_offset;
				} else {
					match_end = (void *) 
						((struct nfhp_rule *) 
						 pt_rule)->e 
						+ ((struct nfhp_rule *) 
						   pt_rule)->e->next_offset;
				}
			} else {
				match = FIRST_IPT_MATCH(&pt_rule->r);
				match_end = IPT_MATCH_END(&pt_rule->r);
			}
			while ((void *) match < match_end){
			       	if (print_match(match , &ip, FMT_PRINT_RULE & 
						(FMT_NUMERIC &&
						 (options & 
						  (OPT_NUMERIC | OPT_DEBUGN))))
				    != 0)
					break;
				match = (void *) match + match->u.match_size;
			}
		}
		if (target) {
			if (target->print)
				target->print(&ip, t, 
					      FMT_PRINT_RULE & 
					      (FMT_NUMERIC && 
					       (options & 
						(OPT_NUMERIC | OPT_DEBUGN))));
		} else if (t && t->u.target_size != sizeof(*t))
			printf("[%u bytes of unknown target data] ",
			       (unsigned int)(t->u.target_size - sizeof(*t)));
	}

	if (options & OPT_DEBUGN || options & OPT_DEBUG) {
		printf("||");
		for (i = 0; i < NUMBER_OF_DIM; i++) {
			if (rule_options & dimid_to_option(i)) {
				printf(" Dim%d:%s%u", i, rule[i].invert ? 
				       "!" : "", rule[i].left);
				if (rule[i].left != rule[i].right)
					printf("-%u", rule[i].right);
			}
		}
	}

	printf("\n");
	
	
}

static void
print_header(const char *chain, int policy)
{
	printf("Chain %s ", chain);
	if (policy == TARGET_ACCEPT)
		printf("(policy %s)\n", IPTC_LABEL_ACCEPT);
	else if (policy == TARGET_DROP)
		printf("(policy %s)\n", IPTC_LABEL_DROP);
	else printf("(unknown policy)\n");

	if (options & OPT_LINENUMBERS)
		printf("%-5s ", "num");

	printf("%-9s ", "target");
	printf(" %-4s ", "prot");
	printf("%-3s ", "opt");
	if (options & OPT_VERBOSE) {
		printf(" %-6s ", "in");
		printf(" %-6s ", "out");
	}
	printf(" %-19s ", "source");
	printf(" %-19s ", "destination");
	printf("\n");
}

static int 
process_list(const void *data, int len)
{
	static u_int32_t rulenum = 0;
	static u_int32_t chain_rulenum = 0;
	const struct nfhp_list_chain *current_chain;
	const void *p = data;

	if (listing_start) {
		current_chain = (struct nfhp_list_chain *) p;
		rulenum = 0;
		chain_rulenum = current_chain->rule_num;
		listing_start = 0;
		p += sizeof(struct nfhp_list_chain);
		print_header(current_chain->label,
			     current_chain->policy);
	}
	while (p < data + len) {
		if (rulenum < chain_rulenum) {
			int rule_size =
				IPT_ALIGN(((struct nfhp_list_rule *) p)->r.size
					  + offsetof(struct nfhp_list_rule,
						     r));
			if (len - (p - data) < rule_size) {
				fprintf(stderr, "%s: incomplete rule\n",
					__FUNCTION__);
				return -1;
			}
			print_rule((struct nfhp_list_rule *) p, 0);
			p += rule_size;
			rulenum++;
		} else {
			if (len - (p - data) <
			    sizeof(struct nfhp_list_chain)) {
				fprintf(stderr, "%s: incomplete chain\n",
					__FUNCTION__);
				return -1;
			}
			printf("\n");
			current_chain = (struct nfhp_list_chain *) p;
			rulenum = 0;
			chain_rulenum = current_chain->rule_num;
			p += sizeof(struct nfhp_list_chain);
			print_header(current_chain->label,
				     current_chain->policy);
			
		}
	}
	return 0;
}

static int
handle_error(int ret, int err)
{
	if (ret == 0 && err == 0) {
		return 0;
	}

	switch (ret) {
	case 0:
		fprintf(stderr, "%s\n", nlhp_error(err));
		break;
	case -1:
		perror("");
		break;
	case -2:
		fprintf(stderr, "Read timeout occured\n");
		break;
	case -3:
		fprintf(stderr, "Not enough memory available\n");
		break;
	case -4:
		fprintf(stderr, "List handler failed\n");
		break;
	}
	return ret;
}

int
do_command(int argc, char *argv[])
{
	int c, invert = 0, ipt_match_num = 0, cmd_return, ret = 1, err;
	struct iptables_target *ipt_target = NULL;
	u_int32_t ipt_matches_size = 0;
	char *protocol = NULL;
	char jumpto[HIPAC_CHAIN_NAME_MAX_LEN] = "";
	struct nfhp_cmd* message = NULL;
	struct dim_match rule[NUMBER_OF_DIM];
	struct in_addr *s_ips = NULL, *d_ips = NULL;
	unsigned int ns_addrs = 0, nd_addrs = 0;
	struct iptables_target *t;
	struct iptables_match *m;
	struct iptables_rule_match *matches = NULL;
	struct iptables_rule_match *matchp;
		
        /* re-set optind to 0 in case do_command gets called
	 * a second time */
	optind = 0;

	/* clear mflags in case do_command gets called a second time
	 * (we clear the global list of all matches for security)*/
	for (m = iptables_matches; m; m = m->next) {
		m->mflags = 0;
	}
	for (t = iptables_targets; t; t = t->next) {
		t->tflags = 0;
		t->used = 0;
	}
	memset(rule, 0, sizeof(struct dim_match) * NUMBER_OF_DIM);
	message = nlhp_new_cmd(CMD_NONE, NULL);
	if (!message)
		perror("nf-hipac: low memory");
	
	message->rule.r.action = TARGET_NONE;
        
        /* Suppress error messages: we may add new options if we
           demand-load a protocol. */
	opterr = 0;
	while ((c = getopt_long (argc, argv, 
				 "-A:D:I:R:L::F::Z::N:X::E:P:t:m:s:d:p:i:j:c:o:fvnxVh::",
				 opts, NULL)) != -1){
		switch(c){
		case 'A':
			add_command(message, CMD_APPEND, invert);
			parse_chain(optarg, message->chain.label);
			break;
		case 'D':
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!') {
				add_command(message, CMD_DELETE_POS, invert);
				parse_chain(optarg, message->chain.label);
				parse_rulenumber(argv[optind++],
						 &message->rule.r.pos);
			} else {
				add_command(message, CMD_DELETE_RULE, invert);
				parse_chain(optarg, message->chain.label);
			}
			break;
		case 'R':
			add_command(message, CMD_REPLACE, invert);
			parse_chain(optarg, message->chain.label);
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				parse_rulenumber(argv[optind++],
						 &message->rule.r.pos);
			else
				exit_error(PARAMETER_PROBLEM,
					   "-%c requires a rule number",
					   cmd2char(CMD_REPLACE));
			break;
		case 'I':
			add_command(message, CMD_INSERT, invert);
			parse_chain(optarg, message->chain.label);
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				parse_rulenumber(argv[optind++],
						 &message->rule.r.pos);
			else message->rule.r.pos = 1;
			break;
		case 'L':
			add_command(message, CMD_LIST, invert);
			if (optarg) parse_chain(optarg, message->chain.label);
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				parse_chain(argv[optind++], 
					    message->chain.label);
			break;
		case 'F':
			add_command(message, CMD_FLUSH, invert);
			if (optarg) parse_chain(optarg, message->chain.label);
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				parse_chain(argv[optind++],
					    message->chain.label);
			break;
		case 'N':
			add_command(message, CMD_NEW_CHAIN, invert);
			if (optarg && (*optarg == '-' || *optarg == '!'))
				exit_error(PARAMETER_PROBLEM,
					   "chain name not allowed to start "
					   "with `%c'\n", *optarg);
			parse_chain(optarg, message->chain.label);
			if (strcmp(optarg, IPTC_LABEL_ACCEPT) == 0
			    || strcmp(optarg, IPTC_LABEL_DROP) == 0
			    || strcmp(optarg, IPTC_LABEL_QUEUE) == 0
			    || strcmp(optarg, IPTC_LABEL_RETURN) == 0
			    || find_target(optarg, TRY_LOAD))
				exit_error(PARAMETER_PROBLEM,
					   "chain name may not clash "
					   "with target name\n");
			break;
		case 'X':
			add_command(message, CMD_DELETE_CHAIN, invert);
			if (optarg) parse_chain(optarg, message->chain.label);
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				parse_chain(argv[optind++], 
					    message->chain.label);
			break;
		case 'E':
			add_command(message, CMD_RENAME_CHAIN, invert);
			parse_chain(optarg, message->chain.label);
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!'){
				parse_chain(argv[optind++], 
					    message->chain.newlabel);
				if (strcmp(message->chain.newlabel, 
					   IPTC_LABEL_ACCEPT) == 0
				    || strcmp(message->chain.newlabel,
					      IPTC_LABEL_DROP) == 0
				    || strcmp(message->chain.newlabel,
					      IPTC_LABEL_QUEUE) == 0
				    || strcmp(message->chain.newlabel,
					      IPTC_LABEL_RETURN) == 0
				    || find_target(message->chain.newlabel, 
						   TRY_LOAD))
					exit_error(PARAMETER_PROBLEM,
						   "chain name may not clash "
						   "with target name\n");
			} else
				exit_error(PARAMETER_PROBLEM,
				           "-%c requires old-chain-name and "
					   "new-chain-name",
					    cmd2char(CMD_RENAME_CHAIN));
		
			break;
		case 'P':
			add_command(message, CMD_SET_POLICY, invert);
			parse_chain(optarg, message->chain.label);
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				parse_policy(argv[optind++],
					     &message->chain.policy);
			else
				exit_error(PARAMETER_PROBLEM,
					   "-%c requires a chain and a policy",
					   cmd2char(CMD_SET_POLICY));
			break;
		case 'h':
			if (!optarg)
				optarg = argv[optind];
			if ((optarg ) && (strcasecmp(optarg, "icmp") == 0))
				print_icmptypes();
			exit_printhelp(matches);
                
		/* Options */	
		case 's':
			check_inverse(optarg, &invert, &optind, argc);
			hbo_parse_hostnetworkmask(argv[optind-1],
						  DIMID_SRC_IP, rule,
						  &s_ips, &ns_addrs);
		        rule[DIMID_SRC_IP].invert = invert;
			break;

		case 'd':
			check_inverse(optarg, &invert, &optind, argc);
			hbo_parse_hostnetworkmask(argv[optind-1],
						  DIMID_DEST_IP, rule,
						  &d_ips, &nd_addrs);
	                rule[DIMID_DEST_IP].invert = invert;
			break;
		case 'p':
			check_inverse(optarg, &invert, &optind, argc);
			for (protocol = argv[optind-1]; *protocol; protocol++)
				*protocol = tolower(*protocol);
			parse_native_protocol(argv[optind-1], rule);
			rule[DIMID_PROTO].invert = invert;
			if (!invert && 
			    (rule[DIMID_PROTO].left ==
			     rule[DIMID_PROTO].right)){
				message->rule.e->ip.proto = 
					rule[DIMID_PROTO].left;
			}
			break;
		case 'f':
			set_option(&options, 0, DIMID_FRAGMENT);
			if (invert){
				rule[DIMID_FRAGMENT].left =
					rule[DIMID_FRAGMENT].right = 0;
			} else {
				rule[DIMID_FRAGMENT].left = 1;
				rule[DIMID_FRAGMENT].right = 65535;
			}
			break;
		case 'j':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "cannot have ! before -%s",
					   optflags[OPT_VERBOSE]);
			set_option(&options, OPT_JUMP, 0);
			parse_target(optarg, jumpto);
			if (strcmp(optarg, IPTC_LABEL_ACCEPT) == 0){
				message->rule.r.action = TARGET_ACCEPT;
				break;
			} else if (strcmp(optarg, IPTC_LABEL_DROP) == 0){
				message->rule.r.action = TARGET_DROP;
				break;
			} else if (strcmp(optarg, IPTC_LABEL_RETURN) == 0){
				message->rule.r.action = TARGET_RETURN;
				break;
			} 
			ipt_target = find_target(jumpto, TRY_LOAD);
			if (ipt_target) {
				size_t size;
				size = IPT_ALIGN(sizeof(struct ipt_entry_target))
					+ ipt_target->size;

				ipt_target->t = fw_calloc(1, size);
				ipt_target->t->u.target_size = size;
				strcpy(ipt_target->t->u.user.name, jumpto);
#ifndef IPTABLES_1_2
				set_revision(ipt_target->t->u.user.name,
					     ipt_target->revision);
#endif			      
				if (ipt_target->init != NULL)
					ipt_target->init(ipt_target->t,
							 &message->rule.e
							 ->nfcache);
				opts = merge_options(opts,
						     ipt_target->extra_opts, 
						     &ipt_target->option_offset);
				message->rule.r.action = TARGET_EXEC;
			} else {
				message->rule.r.action = TARGET_CHAIN;
			}
			break;
		case 'i':
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, 0, DIMID_INIFACE);
			parse_native_interface(argv[optind-1], 
					       message->rule.indev);
			rule[DIMID_INIFACE].invert = invert;
			break;
		case 'o':
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, 0, DIMID_OUTIFACE);
			parse_native_interface(argv[optind-1], 
					       message->rule.outdev);
			rule[DIMID_OUTIFACE].invert = invert;
			break;
		case 'v':
			set_option(&options, OPT_VERBOSE, 0);
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "cannot have ! before -%s",
					   optflags[OPT_VERBOSE]);
			break;
		case 'V':	
			printf("%s v%s\n",
			       program_name, program_version);
			exit(0);
		case 'n':
			set_option(&options, OPT_NUMERIC, 0);
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "cannot have ! before -%s",
					   optflags[OPT_NUMERIC]);
			break;
		case 't':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag before --table");
			if (strcmp(argv[optind-1], "filter") != 0)
				exit_error(PARAMETER_PROBLEM,
					   "no other table than "
					   "filter supported");
			break;
		case 'm': 
		{
			size_t size;
			int i, found = 0;
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag before --match");
			for (i = 0; i < HIPAC_NATIVE_MATCHES; i++){
				if (strcmp(optarg, native_matches[i]) == 0){
					found = 1;
					break;
				}
			}
			if (found)
				break;
			m = find_match(optarg, LOAD_MUST_SUCCEED, &matches);
			size = IPT_ALIGN(sizeof(struct ipt_entry_match))
					 + m->size;
			m->m = (void *) message->rule.e
				+ IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size;
			ipt_matches_size += size;
			m->m->u.match_size = size;
			strcpy(m->m->u.user.name, m->name);
#ifndef IPTABLES_1_2
			set_revision(m->m->u.user.name, m->revision);
#endif
			if (m->init != NULL)
				m->init(m->m, &message->rule.e->nfcache);
			opts = merge_options(opts, m->extra_opts,
					     &m->option_offset);
		}
			break;
		case 9:
			// --debug matched
			set_option(&options, OPT_DEBUG, 0);
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "cannot have ! before -%s",
					   optflags[OPT_DEBUG]);
			break;
		case 10:
			// --debugn matched
			set_option(&options, OPT_DEBUGN, 0);
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "cannot have ! before -%s",
					   optflags[OPT_DEBUGN]);
			break;
			
		case 11:	
			// --line matched
			set_option(&options, OPT_LINENUMBERS, 0);
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "cannot have ! before -%s",
					   optflags[OPT_LINENUMBERS]);
			break;
		case 12:
			// --state matched
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, 0, DIMID_STATE);
			parse_states(argv[optind-1], rule);
			rule[DIMID_STATE].invert = invert;
			break;
		case 14:
			// --ttl matched
			check_inverse(optarg, &invert, &optind, argc);
			if (invert)
				exit_error(PARAMETER_PROBLEM, 
					   "no ! with --ttl allowed");
			parse_ttls(argv[optind-1], rule);
			break;
		case 15:
			// --icmp-type matched
			check_inverse(optarg, &invert, &optind, argc);
			if (invert)
				exit_error(PARAMETER_PROBLEM, 
					   "no ! with --icmp-type allowed");
			set_option(&options, 0, DIMID_ICMP_TYPE);
			parse_icmp(argv[optind-1], rule);
			break;
		case 16:
                        // --source-port matched
			if (rule[DIMID_PROTO].invert)
				exit_error(PARAMETER_PROBLEM,
					   "--sport only works with " 
					   "protocols tcp or udp, so you can't"
					   " use ! on the protocol field");
			check_inverse(optarg, &invert, &optind, argc);
			if ((rule[DIMID_PROTO].left == IPPROTO_TCP  &&
			     rule[DIMID_PROTO].right == IPPROTO_TCP) ||
			    (rule[DIMID_PROTO].left == IPPROTO_UDP &&
			     rule[DIMID_PROTO].right == IPPROTO_UDP)){
				parse_ports(argv[optind-1], 
					    DIMID_SPORT, rule);
				rule[DIMID_SPORT].invert = invert;
			} else exit_error(PARAMETER_PROBLEM,
					  "Can't use --sport without "
					  "protocol. You have to specify " 
					  "protocol (tcp or udp) first");
				
			break;
		case 17:
			// --dest-port matched
			if (rule[DIMID_PROTO].invert)
				exit_error(PARAMETER_PROBLEM,
					   "--dport only works with " 
					   "protocols tcp or udp, so you can't"
					   " use ! on the protocol field");
			check_inverse(optarg, &invert, &optind, argc);
			if ((rule[DIMID_PROTO].left == IPPROTO_TCP &&
			    rule[DIMID_PROTO].right == IPPROTO_TCP) ||
			    (rule[DIMID_PROTO].left == IPPROTO_UDP &&
			     rule[DIMID_PROTO].right == IPPROTO_UDP)){
				parse_ports(argv[optind-1], 
					    DIMID_DPORT, rule);
				rule[DIMID_DPORT].invert = invert;
			} else exit_error(PARAMETER_PROBLEM,
					  "Can't use --dport without "
					  "protocol. You have to specify " 
					  "protocol (tcp or udp) first");
			break;
		case 18:
			// --syn matched
			check_inverse(optarg, &invert, &optind, argc);
			if (invert)
				exit_error(PARAMETER_PROBLEM, 
					   "no ! with --syn allowed");	
			set_option(&options, 0, DIMID_SYN);
			parse_syn("SYN", rule);
			break;
		case 19:
			// --not-syn matched
			check_inverse(optarg, &invert, &optind, argc);
			if (invert)
				exit_error(PARAMETER_PROBLEM, 
					   "no ! with --not-syn allowed");
			set_option(&options, 0, DIMID_SYN);
			parse_syn("NOT-SYN", rule);
			break;
		case 1:	
			if (optarg[0] == '!' && optarg[1] == '\0') {
				if (invert)
					exit_error(PARAMETER_PROBLEM,
						   "multiple consecutive ! not"
						   " allowed");
				invert = TRUE;
				optarg[0] = '\0';
				continue;
			}
			printf("Bad argument `%s'\n", optarg);
			exit_tryhelp(2);
			break;
		default: /* non option */
			if (!ipt_target
			    || !(ipt_target->parse(c -
						   ipt_target->option_offset,
						   argv, invert,
						   &ipt_target->tflags,
						   message->rule.e,
						   &ipt_target->t))) {
				for (matchp = matches; matchp;
				     matchp = matchp->next) {
					if (matchp->match
					    ->parse(c - matchp->match
						    ->option_offset,
						     argv, invert,
						     &matchp->match->mflags,
						     message->rule.e,
						     &message->rule.e->nfcache,
						     &matchp->match->m))
						break;
				}
				m = matchp ? matchp->match : NULL;
				
				if (!m)
					exit_error(PARAMETER_PROBLEM,
						   "Unknown arg `%s'",
						   argv[optind-1]);
			}


		}
		invert = FALSE;
	}

	for (matchp = matches; matchp; matchp = matchp->next) {
		ipt_match_num++;
		matchp->match->final_check(matchp->match->mflags);
	}

	if (ipt_target)
		ipt_target->final_check(ipt_target->tflags);

	if (optind < argc)
		exit_error(PARAMETER_PROBLEM,
			   "unknown arguments found on commandline");
	
	if (!message->cmd)
		exit_error(PARAMETER_PROBLEM, "no command specified");
	
	generic_opt_check(message->cmd, options);

	if ((ns_addrs > 1) && (rule[DIMID_SRC_IP].invert))
		exit_error(PARAMETER_PROBLEM, "! not allowed with multiple"
			   " source or destination IP addresses");

	if ((nd_addrs > 1) && (rule[DIMID_DEST_IP].invert))
		exit_error(PARAMETER_PROBLEM, "! not allowed with multiple"
			   " source or destination IP addresses");

	if (message->cmd == CMD_REPLACE && (ns_addrs > 1 || nd_addrs > 1))
		exit_error(PARAMETER_PROBLEM, "Replacement rule does not "
			   "specify a unique address");

	if (message->cmd == CMD_APPEND
	    || message->cmd == CMD_DELETE_RULE
	    || message->cmd == CMD_DELETE_POS
	    || message->cmd == CMD_INSERT
	    || message->cmd == CMD_REPLACE) {
		
		
       		if (strcasecmp(message->chain.label, "INPUT") == 0) {
			/* -o not valid with incoming packets. */
			if (options & dimid_to_option(DIMID_OUTIFACE))
				exit_error(PARAMETER_PROBLEM,
					   "Can't use -%s with %s",
					   optflags[NUMBER_OF_OPT 
						    + DIMID_OUTIFACE],
					   message->chain.label);
		}

		if (strcasecmp(message->chain.label, "OUTPUT") == 0) {
			/* -i not valid with outgoing packets */
			if (options & dimid_to_option(DIMID_INIFACE))
				exit_error(PARAMETER_PROBLEM,
					   "Can't use -%s with %s",
					   optflags[NUMBER_OF_OPT + 
						    DIMID_INIFACE],
					   message->chain.label);
		}
		
		if ((options & dimid_to_option(DIMID_SYN))
		    && !((options & dimid_to_option(DIMID_PROTO)) &&
			 (rule[DIMID_PROTO].left == IPPROTO_TCP) &&
			 (rule[DIMID_PROTO].right == IPPROTO_TCP) &&
			 (rule[DIMID_PROTO].invert == 0)))
			exit_error(PARAMETER_PROBLEM,
				   "Can use --syn or --not-syn only "
				   "with protocol tcp");
				
		if ((options & dimid_to_option(DIMID_ICMP_TYPE))
		    && !((options & dimid_to_option(DIMID_PROTO)) &&
			 (rule[DIMID_PROTO].left == IPPROTO_ICMP) &&
			 (rule[DIMID_PROTO].right == IPPROTO_ICMP) &&
			 (rule[DIMID_PROTO].invert == 0))) 
			exit_error(PARAMETER_PROBLEM,
				   "Can use --icmp-type only "
				   "with protocol icmp");
		
		if ((options & dimid_to_option(DIMID_SYN))
		    || (options & dimid_to_option(DIMID_SPORT))
		    || (options & dimid_to_option(DIMID_DPORT))
		    || (options & dimid_to_option(DIMID_ICMP_TYPE))) {
			if (options & dimid_to_option(DIMID_FRAGMENT)) {
				if (rule[DIMID_FRAGMENT].left == 1){
					exit_error(PARAMETER_PROBLEM,
						   "syn, port and "
						   "icmp type matches are "
						   "not allowed on "
						   "fragments");
				}
			} else {
				set_option(&options, 0, DIMID_FRAGMENT);
				rule[DIMID_FRAGMENT].left = 0;
				rule[DIMID_FRAGMENT].right = 0;
			}
		}
	}
	
	if (message->cmd != CMD_LIST){
		int i, j;
		int cmd_size;
		struct nfhp_cmd* new_message = NULL;
		if (HAS_CHAIN_TARGET(&message->rule.r)){
			strcpy((void *) message->rule.e	
			       + IPT_ALIGN(sizeof(struct ipt_entry))
			       + ipt_matches_size, jumpto);
			cmd_size = sizeof(struct nfhp_cmd)
				+ IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size
				+ strlen(jumpto) + 1;
			message->rule.e->target_offset =
				IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size;
			message->rule.e->next_offset =
				IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size
				+ strlen(jumpto) + 1;
		} else if (HAS_IPT_TARGET(&message->rule.r)){
			memcpy((void *) message->rule.e	
			       + IPT_ALIGN(sizeof(struct ipt_entry))
			       + ipt_matches_size, 
			       ipt_target->t, ipt_target->t->u.target_size);
			cmd_size = sizeof(struct nfhp_cmd)
				+ IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size
				+ ipt_target->t->u.target_size;
			message->rule.e->target_offset =
				IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size;
			message->rule.e->next_offset =
				IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size
				+ ipt_target->t->u.target_size;
			free(ipt_target->t);
		} else if (ipt_match_num){
			cmd_size =  sizeof(struct nfhp_cmd)
				+ IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size;
			message->rule.e->target_offset =
				IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size;
			message->rule.e->next_offset =
				IPT_ALIGN(sizeof(struct ipt_entry))
				+ ipt_matches_size;
		} else {
			cmd_size = sizeof(struct nfhp_cmd);
		}

		message->rule.r.match_offset = ipt_matches_size;

		if ((message->cmd == CMD_DELETE_RULE) 
		    && HAS_IPT_ENTRY(&message->rule.r)){
			u_int16_t *cmp;
			struct ipt_entry_match *k;
			cmp = (void *) message->rule.e
				+ message->rule.e->next_offset;
			k = (void *) message->rule.e
				+ IPT_ALIGN(sizeof(struct ipt_entry));
			cmd_size += ipt_match_num * (sizeof(u_int16_t));
			for (i = 0; i < ipt_match_num; i++){
				for (matchp = matches; matchp;
				     matchp = matchp->next) {
					//if (!matchp->match->used)
					//continue;
					//fixme: NULL AHNUNG
					if (strcmp(k->u.user.name,
						   matchp->match->name) == 0){
						cmp[i] = matchp->
							match->userspacesize;
						k = NEXT_IPT_MATCH(k);
						break;
					}
				}
			}
			if (HAS_IPT_TARGET(&message->rule.r)){
				cmp[i] = ipt_target->userspacesize;
				cmd_size += (sizeof(u_int16_t));
			}
		}
		
		if ((options & OPT_DEBUG) || 
		    (options & OPT_DEBUGN)){
			print_header(message->chain.label, 
				     message->chain.policy);
		}
		
		if (ns_addrs < 1) ns_addrs = 1;
		if (nd_addrs < 1) nd_addrs = 1;
		for (i = 0; i < ns_addrs; i++)
			for (j = 0; j < nd_addrs; j++){
				if (s_ips)
					rule[DIMID_SRC_IP].left =
						rule[DIMID_SRC_IP].right =
						(s_ips[i]).s_addr;
				if (d_ips)
					rule[DIMID_DEST_IP].left = 
						rule[DIMID_DEST_IP].right = 
						(d_ips[j]).s_addr;
				
				fill_hipac_matches(message, rule);
				
				if (((i + 1) * (j + 1))
					< (ns_addrs * nd_addrs)){
					new_message = 
						nlhp_copy_cmd(message,
							      cmd_size);
					if (!new_message)
						perror("nf-hipac: low memory");
				}

				if ((options & OPT_DEBUG) || 
				    (options & OPT_DEBUGN)){
					print_rule((struct nfhp_list_rule *)
						   &(message->rule), 1);
					nlhp_free_cmd(message);
				} else {
					if (options & OPT_VERBOSE)
						print_rule(
							(struct 
							 nfhp_list_rule *)
							&(message->rule), 1);
					cmd_return = nlhp_send_cmd(message,
								   cmd_size,
								   NULL, &err);
					nlhp_free_cmd(message);
					ret = handle_error(cmd_return, err);
					if (ret != 0)
						return ret;
				}
				message = new_message;
			}
	} else {
		listing_start = 1;
		cmd_return = nlhp_send_cmd(message, sizeof(struct nfhp_cmd),
					   process_list, &err);
		ret = handle_error(cmd_return, err);
	}

	clear_rule_matches(&matches);

	if (s_ips) 
		free(s_ips);
	if (d_ips)
		free(d_ips);

	free_opts(1);
	
	options = 0;

	return ret;
}

